import re
import json
import multiprocessing as mp
import time
import traceback
import sys
import six.moves.queue as Queue
import six
import wx
import wx.py.dispatcher as dp
import wx.lib.agw.aui as aui
from ..auibarpopup import AuiToolBarPopupArt
from . import graph
from .bsmxpm import module_xpm, switch_xpm, in_xpm, out_xpm, inout_xpm,\
                    module_grey_xpm, switch_grey_xpm, in_grey_xpm,\
                    out_grey_xpm, inout_grey_xpm, step_svg, step_grey_svg, run_svg, run_grey_svg, \
                    pause_svg, pause_grey_svg, setting_xpm, radio_disabled_svg, \
                    radio_activated_svg, radio_checked_svg, radio_unchecked_svg
from .simprocess import sim_process, SC_OBJ_UNKNOWN, SC_OBJ_SIGNAL, SC_OBJ_INPUT,\
                        SC_OBJ_OUTPUT, SC_OBJ_INOUT, SC_OBJ_CLOCK, SC_OBJ_XSC_PROP,\
                        SC_OBJ_XSC_ARRAY_ITEM, SC_OBJ_MODULE, SC_OBJ_XSC_ARRAY
from .. import propgrid as pg
from .pymgr_helpers import Gcm
from .autocomplete import AutocompleteTextCtrl
from .utility import MakeBitmap, FastLoadTreeCtrl, PopupMenu
from .. import to_byte
from .utility import svg_to_bitmap

Gcs = Gcm()


class Simulation(object):
    def __init__(self, parent, num=None):
        if num is not None and isinstance(num, int) and \
           num not in Gcs.get_all_managers():
            self.num = num
        else:
            self.num = Gcs.get_next_num()
        Gcs.set_active(self)
        self.frame = parent
        # simulation filename
        self.filename = ""
        # the dict holds all the objects from the simulation
        self.objects = None
        # create the simulation kernel
        self._cmd_id = 0
        self.qresp = None
        self.qcmd = None
        self.sim_process = None
        self.status = {'running': False, 'valid': False}
        self._interfaces = {}

    def release(self):
        """destroy simulation"""
        self.frame = None
        self.stop()
        Gcs.destroy(self.num)

    def __getitem__(self, obj):
        """read the object"""
        return self.read(objects=obj)

    def __setitem__(self, obj, value):
        """write the value to the object"""
        if isinstance(obj, six.string_types):
            return self.write({obj: value})
        elif isinstance(obj, (list, tuple)):
            return self.write(dict(zip(obj, value)))
        else:
            raise ValueError()

    def _send_command(self, cmd, **kwargs):
        """
        send the command to the simulation process

        don't call this function directly unless you know what it is doing.
        """
        try:
            # always increase the command ID
            cid = self._cmd_id
            self._cmd_id += 1

            # return, if the previous call has not finished
            # it may happen when the previous command is waiting for response,
            # and another command is sent (by clicking a button)
            if self.qresp is None or self.qcmd is None or self.sim_process is None:
                raise KeyboardInterrupt
            block = kwargs.get('block', True)

            if not kwargs.get('silent', True):
                print(cmd, cid, kwargs)

            self.qcmd.put({'id': cid, 'cmd': cmd, 'arguments': kwargs})
            if block is True:
                # wait for the command to finish
                while True:
                    resp = self.qresp.get()
                    rtn = self._process_response(resp)
                    if resp.get('id', -1) == cid:
                        return rtn
            return True
        except:
            traceback.print_exc(file=sys.stdout)

    def is_valid(self):
        return self.status['valid']

    def is_running(self):
        return self.status['running']

    def wait_until_simulation_paused(self, timeout=None):
        """
        wait for the simulation to pause

        return: It will not return until the simulation has paused.
                True: the simulation is valid and paused;
                False: otherwise.
        timeout: the maximum waiting period (in second)
        """
        # return if the simulation is not valid
        if self.qresp is None or self.qcmd is None or self.sim_process is None:
            return False
        start = -1
        if timeout is not None:
            start = time.time()
        while True:
            resp = self.get_status()
            if not resp['running']:
                return True
            wx.YieldIfNeeded()
            time.sleep(0.1)
            if start > 0 and time.time() - start > timeout:
                return False
        return False

    def start(self):
        """create an empty simulation"""
        self.stop()
        self.qresp = mp.Queue()
        self.qcmd = mp.Queue()
        self.sim_process = mp.Process(target=sim_process,
                                      args=(self.qresp, self.qcmd))
        self.sim_process.start()
        self._interfaces = self._send_command('get_interfaces')
        if self._set_interfaces:
            self._set_interfaces(self._interfaces)

    def _set_interfaces(self, interfaces):
        for item in interfaces:
            if hasattr(self, item):
                continue

            def wrapper(cmd, **kwargs):
                return self._send_command(cmd, **kwargs)

            formatted_args = interfaces[item]['args']
            formatted_args = formatted_args.lstrip('(').rstrip(')')
            formatted_args2 = []
            for arg in formatted_args.split(','):
                if '=' in arg:
                    # remove the default value
                    arg = arg[:arg.find('=')]
                    arg = "{0}={0}".format(arg)
                formatted_args2.append(arg)

            fndef = 'lambda %s: wrapper("%s", %s)' % (
                formatted_args, item, ','.join(formatted_args2))
            fake_fn = eval(fndef, {'wrapper': wrapper, 'item': item})
            fake_fn.__doc__ = interfaces[item]['doc']
            setattr(self, item, fake_fn)

    def load(self, filename=None, block=True):
        """
        load the simulation library (e.g., dll)

        if filename is None, show a file open dialog to select and open the
        simulation
        """
        if filename is None:
            style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
            dlg = wx.FileDialog(self.frame, "Choose a file", "", "", "*.*",
                                style)
            if dlg.ShowModal() == wx.ID_OK:
                filename = dlg.GetPath()
            dlg.Destroy()

        if filename is None:
            return False

        self.start()
        self.filename = filename
        self._send_command('load', block=block, filename=filename)
        return self.get_status()

    def run(self, to=None, more=None, block=False):
        """
        keep running the simulation

        The simulation is executed step by step. After each step, the simulation
        'server' will notify the 'client' to update the GUI.
        """
        self.set_parameter(to=to, more=more, block=False)
        return self.step(running=True, block=block)

    def stop(self):
        """destroy the simulation"""
        if self.qresp is None or self.qcmd is None or self.sim_process is None:
            return
        # stop the simulation kernel. No block operation allowed since
        # no response from the subprocess
        self._send_command('exit', block=False)
        while not self.qresp.empty():
            self.qresp.get_nowait()
        self.sim_process.join()
        self.sim_process = None
        # stop the client
        self._process_response({'cmd': 'exit'})

    def reset(self):
        """reset the simulation"""
        if self.filename:
            return self.load(self.filename)
        return False

    def _object_list(self, objs):
        """help function to generate object list"""
        if isinstance(objs, six.string_types):
            return [objs]
        elif isinstance(objs, (list, tuple)):
            return objs
        else:
            raise ValueError()

    def global_object_name(self, obj):
        """generate the global name for simulation object (num.name)"""
        if obj in self.objects:
            return "%d." % self.num + obj
        return None

    def monitor(self, objects, grid=None, index=-1):
        """
        show the register in a propgrid window

        grid: If the grid is None, the objects will be added to the active propgrid
        window. If the active propgrid window is also None, it will create one.

        index: the index of the object in the propgrid window. -1 to append.
        """
        if not objects:
            return None
        if grid is None:
            grid = SimPropGrid.GCM.get_active()
        if not grid:
            grid = sim.propgrid()
        if not grid:
            return None
        props = []
        for name in self._object_list(objects):
            obj = self.objects.get(name, None)
            # ignore the invalid object
            if obj is None:
                print("invalid object: ", name)
                continue
            if obj['kind'] == 'sc_module':
                prop = grid.InsertSeparator(
                    self.global_object_name(obj['name']), obj['basename'],
                    index)
            else:
                prop = grid.InsertProperty(
                    self.global_object_name(obj['name']), obj['basename'],
                    obj['value'], index)
                prop.SetGripperColor(self.frame.GetColor())
                if not obj['writable']:
                    prop.SetControlStyle('none')
                prop.SetShowCheck(obj['readable'])
                # set value invalid since obj['value'] may have garbage value
                # (not read the actual value from simulation yet.);
                # will set it to true once the value is updated.
                prop.SetValueValid(False)
            props.append(prop)
            if index != -1:
                index += 1
        if len(props) == 1:
            return props[0]
        return props

    def _process_response(self, resp):
        """process the response from the simulation core"""
        try:
            wx.CallAfter(dp.send,
                         signal='sim.response',
                         sender=self,
                         resp=resp)
            command = resp.get('cmd', '')
            value = resp.get('value', False)
            args = resp.get('arguments', {})
            if command == 'load':
                self.objects = value
                self.filename = args['filename']
            elif command in ['read', 'read_buf', 'write']:
                # single value, return the value, not the dict
                if len(value) == 1:
                    value = list(value.values())[0]
            elif command == 'get_status':
                running = self.status['running']
                if running and not value.get('running', running):
                    # if the queues are almost empty, retrieve the data
                    self.time_stamp(insecond=False, block=False)
                    self.read(objects=[], block=False)
                    self.read_buf([], block=False)
                self.status.update(value)
            return value
        except:
            traceback.print_exc(file=sys.stdout)

    def process_response(self):
        if not self.qresp:
            return None
        try:
            # process the response
            resp = self.qresp.get_nowait()
            if resp:
                return self._process_response(resp)
        except Queue.Empty:
            pass

        if self.status['running'] and self.qresp.empty() and self.qcmd.empty():
            # if the queues are almost empty, retrieve the data
            self.time_stamp(insecond=False, block=False)
            self.read(objects=[], block=False)
            self.read_buf([], block=False)

        return None


class ModuleTree(FastLoadTreeCtrl):
    """the tree control to show the hierarchy of the objects in the simulation"""
    def __init__(self, parent, style=wx.TR_DEFAULT_STYLE):
        style = style | wx.TR_HAS_VARIABLE_ROW_HEIGHT | wx.TR_HIDE_ROOT |\
                wx.TR_MULTIPLE | wx.TR_LINES_AT_ROOT
        FastLoadTreeCtrl.__init__(self, parent, self.get_children, style=style)
        imglist = wx.ImageList(16, 16, True, 10)
        for xpm in [
                module_xpm, switch_xpm, in_xpm, out_xpm, inout_xpm,
                module_grey_xpm, switch_grey_xpm, in_grey_xpm, out_grey_xpm,
                inout_grey_xpm
        ]:
            imglist.Add(wx.Bitmap(to_byte(xpm)))
        self.AssignImageList(imglist)
        self.objects = None

    def get_children(self, item):
        """ callback function to return the children of item """
        if item == self.GetRootItem():
            parent = ""
        else:
            ext = self.GetExtendObj(item)
            parent = ext['name']
        children = []
        for key, obj in six.iteritems(self.objects):
            if obj['nkind'] != SC_OBJ_UNKNOWN and obj['parent'] == parent:
                child = {'label': obj['basename']}
                nkind = obj['nkind']
                img = [-1, -1]
                if nkind == SC_OBJ_MODULE:
                    img = [0, 0]
                elif nkind in [
                        SC_OBJ_SIGNAL, SC_OBJ_CLOCK, SC_OBJ_XSC_PROP,
                        SC_OBJ_XSC_ARRAY, SC_OBJ_XSC_ARRAY_ITEM
                ]:
                    img = [1, 1]
                elif nkind == SC_OBJ_INPUT:
                    img = [2, 2]
                elif nkind == SC_OBJ_OUTPUT:
                    img = [3, 3]
                elif nkind == SC_OBJ_INOUT:
                    img = [4, 4]
                else:
                    raise ValueError('Invalid object type')
                child['img'] = img[0]
                child['imgsel'] = img[1]
                child['data'] = key
                # add the sign to append the children later
                child['is_folder'] = nkind in [SC_OBJ_MODULE, SC_OBJ_XSC_ARRAY]
                children.append(child)
        return children

    def OnCompareItems(self, item1, item2):
        """compare the two items for sorting"""
        def SortByName(obj1, obj2):
            """compare the two items based on its type and name"""
            type1 = obj1['nkind'] in [SC_OBJ_MODULE, SC_OBJ_XSC_ARRAY]
            type2 = obj2['nkind'] in [SC_OBJ_MODULE, SC_OBJ_XSC_ARRAY]
            if type1 == type2:
                if obj1['name'] > obj2['name']:
                    return 1
            elif type1 < type2:
                return 1
            return -1

        data1 = self.GetExtendObj(item1)
        data2 = self.GetExtendObj(item2)
        rtn = -2
        if data1 and data2:
            return SortByName(data1, data2)
        return rtn

    def Load(self, objects):
        """load the new simulation"""
        self.objects = objects
        self.FillTree()

    def FillTree(self):
        """fill the simulation objects tree"""
        #clear the tree control
        self.DeleteAllItems()
        if self.objects is None:
            return

        # add the root item
        item = self.AddRoot("bsmedit")
        # fill the top level item
        self.FillChildren(item)
        # any item? expand it
        item, _ = self.GetFirstChild(item)
        if item.IsOk():
            self.Expand(item)

    def GetExtendObj(self, item):
        """return the extend object"""
        data = self.GetItemData(item)
        if data:
            return self.objects.get(data, None)
        return None

    def FindItem(self, parent, name):
        """find the first child of parent by its name"""
        (child, cookie) = self.GetFirstChild(parent)
        while child:
            ext = self.GetExtendObj(child)
            if name == ext['name']:
                return child
            if self.ItemHasChildren(child):
                self.FillNodes(child)
                grandchild = self.FindItem(child, name)
                if grandchild.IsOk():
                    return grandchild
            (child, cookie) = self.GetNextChild(parent, cookie)
        return wx.TreeItemId()

    def SetActiveNode(self, name):
        """select the item by its name"""
        item = self.GetRootItem()
        child = self.FindItem(item, name)
        if child.IsOk():
            self.EnsureVisible(child)
            self.UnselectAll()
            self.SelectItem(child)


class DumpDlg(wx.Dialog):
    def __init__(self, parent, objects, active, tracefile=True):
        wx.Dialog.__init__(self, parent, id=wx.ID_ANY, title="Dump...")

        self.objects = objects
        self.traceFile = tracefile

        szAll = wx.BoxSizer(wx.VERTICAL)

        if self.traceFile:
            sbox = wx.StaticBox(self, label="&File name")
            szFile = wx.StaticBoxSizer(sbox, wx.HORIZONTAL)
            self.tcFile = wx.TextCtrl(sbox, value=active)
            self.tcFile.SetMaxLength(0)
            szFile.Add(self.tcFile, 1, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
            self.btnSelectFile = wx.Button(sbox, label="...", size=(25, -1))
            szFile.Add(self.btnSelectFile, 0,
                       wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
            szAll.Add(szFile, 0, wx.ALL | wx.EXPAND, 5)

        sbox = wx.StaticBox(self, wx.ID_ANY, "&Signal")
        szSignal = wx.StaticBoxSizer(sbox, wx.VERTICAL)
        self.tcSignal = AutocompleteTextCtrl(sbox,
                                             value=active,
                                             completer=self.Completer)
        szSignal.Add(self.tcSignal, 0, wx.ALL | wx.EXPAND, 5)
        self.cbTrigger = wx.CheckBox(sbox, label="Use Trigger Signal")
        szSignal.Add(self.cbTrigger, 0, wx.ALL, 5)
        self.tcValid = AutocompleteTextCtrl(sbox, completer=self.Completer)
        szSignal.Add(self.tcValid, 0, wx.ALL | wx.EXPAND, 5)
        rbTriggerChoices = ["&Pos Edge", "&Neg Edge", "Both Edge"]
        self.rbTrigger = wx.RadioBox(sbox,
                                     label="Trigger",
                                     choices=rbTriggerChoices)
        self.rbTrigger.SetSelection(2)
        szSignal.Add(self.rbTrigger, 0, wx.ALL | wx.EXPAND, 5)
        szAll.Add(szSignal, 1, wx.ALL | wx.EXPAND, 5)

        if self.traceFile:
            rbFormatChoices = [u"&VCD", u"&BSM"]
            self.rbFormat = wx.RadioBox(self,
                                        label="&Format",
                                        choices=rbFormatChoices)
            self.rbFormat.SetSelection(1)
            szAll.Add(self.rbFormat, 0, wx.ALL | wx.EXPAND, 5)
        else:
            szSize = wx.BoxSizer(wx.HORIZONTAL)
            szSize.Add(wx.StaticText(self, wx.ID_ANY, "Size"), 0, wx.ALL, 5)
            self.spinSize = wx.SpinCtrl(self,
                                        style=wx.SP_ARROW_KEYS,
                                        min=1,
                                        max=2**31 - 1,
                                        initial=256)
            szSize.Add(self.spinSize, 1, wx.EXPAND | wx.ALL, 5)
            szAll.Add(szSize, 0, wx.ALL | wx.EXPAND, 5)

        self.m_staticline1 = wx.StaticLine(self, style=wx.LI_HORIZONTAL)
        szAll.Add(self.m_staticline1, 0, wx.EXPAND | wx.ALL, 5)

        szConfirm = wx.BoxSizer(wx.HORIZONTAL)
        self.btnOK = wx.Button(self, wx.ID_OK, u"OK")
        szConfirm.Add(self.btnOK, 0, wx.ALL, 5)
        self.btnCancel = wx.Button(self, wx.ID_CANCEL, u"Cancel")
        szConfirm.Add(self.btnCancel, 0, wx.ALL, 5)
        szAll.Add(szConfirm, 0, wx.ALIGN_RIGHT, 5)

        self.SetSizer(szAll)
        self.Layout()
        szAll.Fit(self)

        self.tcValid.Enable(self.cbTrigger.GetValue())
        # Connect Events
        if self.traceFile:
            self.btnSelectFile.Bind(wx.EVT_BUTTON, self.OnBtnSelectFile)
        self.cbTrigger.Bind(wx.EVT_CHECKBOX, self.OnCheckVal)
        self.btnOK.Bind(wx.EVT_BUTTON, self.OnBtnOK)

        self.trace = {}

    def Completer(self, query):
        """return all the simulation object for auto complete"""
        root = query
        if query.find('.'):
            root = query[0:query.rfind('.') + 1]
        # only report the direct children of 'query', for example
        # query = 'top.si'
        # will return all the children of top, whose name starts with 'si'
        objs = []
        for obj in self.objects:
            if obj.startswith(query):
                item = obj[len(root):]
                if item.find('.') != -1:
                    item = item[:item.find('.')]
                objs.append(item)
        if objs:
            objs = list(set(objs))
            objs.sort()
        return objs, objs, len(query) - len(root)

    def OnBtnSelectFile(self, event):
        wildcard = "BSM Files (*.bsm)|*.bsm|All Files (*.*)|*.*"
        dlg = wx.FileDialog(self,
                            "Select BSM dump file",
                            '',
                            '',
                            wildcard,
                            style=wx.FD_OPEN)
        if dlg.ShowModal() == wx.ID_OK:
            self.tcFile.SetValue(dlg.GetPath())

    def OnCheckVal(self, event):
        self.tcValid.Enable(self.cbTrigger.GetValue())

    def OnBtnOK(self, event):
        self.trace = {}
        if self.traceFile:
            # trace file
            self.trace['filename'] = self.tcFile.GetValue()
            self.trace['format'] = self.rbFormat.GetSelection()
        else:
            # trace buffer
            self.trace['size'] = self.spinSize.GetValue()
        self.trace['signal'] = self.tcSignal.GetValue()
        self.trace['valid'] = None
        if self.cbTrigger.GetValue():
            self.trace['valid'] = self.tcValid.GetValue()
        self.trace['trigger'] = self.rbTrigger.GetSelection()
        event.Skip()

    def GetTrace(self):
        return self.trace


class SimPanel(wx.Panel):
    ID_SIM_STEP = wx.NewId()
    ID_SIM_RUN = wx.NewId()
    ID_SIM_PAUSE = wx.NewId()
    ID_SIM_SET = wx.NewId()
    ID_MP_DUMP = wx.NewId()
    ID_MP_TRACE_BUF = wx.NewId()
    ID_MP_ADD_TO_NEW_VIEWER = wx.NewId()
    ID_MP_ADD_TO_VIEWER_START = wx.NewId()

    def __init__(self, parent, num=None, filename=None, silent=False):
        wx.Panel.__init__(self, parent)

        self.is_destroying = False
        self._color = wx.Colour(178, 34, 34)
        self.toolbarart = AuiToolBarPopupArt(self)
        self.tb = aui.AuiToolBar(self,
                                 -1,
                                 agwStyle=aui.AUI_TB_OVERFLOW
                                 | aui.AUI_TB_PLAIN_BACKGROUND)
        self.tb.SetToolBitmapSize(wx.Size(16, 16))
        xpm2bmp = wx.Bitmap
        self.tb.AddTool(self.ID_SIM_STEP, "Step", svg_to_bitmap(step_svg),
                        svg_to_bitmap(step_grey_svg), wx.ITEM_NORMAL,
                        "Step the simulation")
        self.tb.AddTool(self.ID_SIM_RUN, "Run", svg_to_bitmap(run_svg),
                        svg_to_bitmap(run_grey_svg), wx.ITEM_NORMAL,
                        "Run the simulation")
        self.tb.AddTool(self.ID_SIM_PAUSE, "Pause", svg_to_bitmap(pause_svg),
                        svg_to_bitmap(pause_grey_svg), wx.ITEM_NORMAL,
                        "Pause the simulation")

        self.tb.AddSeparator()

        self.tcStep = wx.SpinCtrlDouble(self.tb,
                                        value='1000',
                                        size=(150, -1),
                                        min=0,
                                        max=1e9,
                                        inc=1)
        self.tcStep.SetDigits(4)
        self.tb.AddControl(wx.StaticText(self.tb, wx.ID_ANY, "Step "))
        self.tb.AddControl(self.tcStep)
        units = ['fs', 'ps', 'ns', 'us', 'ms', 's']
        self.cmbUnitStep = wx.Choice(self.tb,
                                     wx.ID_ANY,
                                     size=(50, -1),
                                     choices=units)
        self.cmbUnitStep.SetSelection(2)
        self.tb.AddControl(self.cmbUnitStep)
        self.tb.AddSeparator()

        self.tcTotal = wx.SpinCtrlDouble(self.tb,
                                         value='-1',
                                         size=(150, -1),
                                         min=-1,
                                         max=1e9,
                                         inc=1)
        self.tcTotal.SetDigits(4)
        self.tb.AddControl(wx.StaticText(self.tb, wx.ID_ANY, "Total "))
        self.tb.AddControl(self.tcTotal)
        self.cmbUnitTotal = wx.Choice(self.tb,
                                      wx.ID_ANY,
                                      size=(50, -1),
                                      choices=units)
        self.cmbUnitTotal.SetSelection(2)
        self.tb.AddControl(self.cmbUnitTotal)
        self.tb.AddStretchSpacer()
        self.tb.AddSimpleTool(self.ID_SIM_SET, "Setting", xpm2bmp(to_byte(setting_xpm)),
                              "Configure the simulation")

        self.tb.SetToolDropDown(self.ID_SIM_SET, True)
        self.tb.SetArtProvider(self.toolbarart)
        self.tb.Realize()
        self.tree = ModuleTree(self)
        self.box = wx.BoxSizer(wx.VERTICAL)
        self.box.Add(self.tb, 0, wx.EXPAND, 5)
        self.box.Add(self.tree, 1, wx.EXPAND)

        self.box.Fit(self)
        self.SetSizer(self.box)

        self.Bind(wx.EVT_TOOL, self.OnProcessCommand)
        self.Bind(wx.EVT_MENU, self.OnProcessCommand)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateCmdUI)
        self.tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnTreeSelChanged)
        self.tree.Bind(wx.EVT_TREE_ITEM_MENU, self.OnTreeItemMenu)
        self.tree.Bind(wx.EVT_TREE_BEGIN_DRAG, self.OnTreeBeginDrag)
        self.Bind(aui.EVT_AUITOOLBAR_TOOL_DROPDOWN,
                  self.OnMenuDropDown,
                  id=self.ID_SIM_SET)
        # simulation
        self.sim = Simulation(self, num)
        if filename is not None or not silent:
            self.sim.load(filename, block=True)
            self.SetParameter()

        dp.connect(receiver=self._process_response,
                   signal='sim.response',
                   sender=self.sim)

        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer, self.timer)
        self.timer.Start(5)

    def OnTimer(self, event):
        try:
            # process the response
            self.sim.process_response()
        except:
            traceback.print_exc(file=sys.stdout)

    def OnMenuDropDown(self, event):
        if event.IsDropDownClicked():
            menu = wx.Menu()
            menu.Append(wx.ID_RESET, "&Reset")
            menu.AppendSeparator()
            menu.Append(wx.ID_EXIT, "&Exit")

            # line up our menu with the button
            tb = event.GetEventObject()
            tb.SetToolSticky(event.GetId(), True)
            rect = tb.GetToolRect(event.GetId())
            pt = tb.ClientToScreen(rect.GetBottomLeft())
            pt = self.ScreenToClient(pt)

            self.PopupMenu(menu, pt)

            # make sure the button is "un-stuck"
            tb.SetToolSticky(event.GetId(), False)

    def Destroy(self):
        """
        Destroy the simulation properly before close the pane.
        """
        dp.disconnect(receiver=self._process_response,
                      signal='sim.response',
                      sender=self.sim)
        self.timer.Stop()
        self.is_destroying = True
        self.sim.release()
        self.sim = None
        super(SimPanel, self).Destroy()

    def SetColor(self, clr):
        self._color = clr

    def GetColor(self):
        return self._color

    def SetParameter(self, block=True):
        """set the simulation parameters with the values from GUI"""
        if not self.sim or not self.sim.is_valid():
            return

        step = self.tcStep.GetValue()
        unitStep = self.cmbUnitStep.GetString(self.cmbUnitStep.GetSelection())
        step = "%f%s" % (step, unitStep)
        total = self.tcTotal.GetValue()
        unitTotal = self.cmbUnitTotal.GetString(
            self.cmbUnitTotal.GetSelection())
        total = "%f%s" % (total, unitTotal)
        self.sim.set_parameter(step=step, total=total, block=block)

    def OnTreeSelChanged(self, event):
        item = event.GetItem()
        if not item.IsOk():
            return
        # pre-read the prop value
        objects = []
        items = self.tree.GetSelections()
        for item in items:
            obj = self.tree.GetExtendObj(item)
            objects.append(obj['name'])
        if objects:
            wx.CallAfter(self.sim.read, objects, block=True)

    def OnTreeItemMenu(self, event):
        item = event.GetItem()
        if not item.IsOk():
            return
        menu = wx.Menu()
        menu.Append(self.ID_MP_DUMP, "&Dump file")
        menu.AppendSeparator()
        menu.Append(self.ID_MP_TRACE_BUF, "&Trace buffer")
        menu.AppendSeparator()
        submenu = wx.Menu()
        submenu.Append(self.ID_MP_ADD_TO_NEW_VIEWER, "&Add to new propgrid")
        submenu.AppendSeparator()
        nid = wx.ID_HIGHEST
        grids = []
        for mag in SimPropGrid.GCM.get_all_managers():
            grids.append(mag)
            submenu.Append(nid, mag.GetLabel())
            nid += 1

        menu.AppendSubMenu(submenu, "Add to...")
        cmd = PopupMenu(self, menu)
        if cmd in [self.ID_MP_DUMP, self.ID_MP_TRACE_BUF]:
            objs = [o for o, v in six.iteritems(self.sim.objects) if \
                    v['numeric'] and v['readable']]
            objs.sort()
            active = ''
            items = self.tree.GetSelections()
            if items:
                active = self.tree.GetExtendObj(items[0])['name']
            dlg = DumpDlg(self, objs, active, cmd == self.ID_MP_DUMP)
            if dlg.ShowModal() == wx.ID_OK:
                t = dlg.GetTrace()
                if cmd == self.ID_MP_DUMP:
                    self.sim.trace_file(t['filename'],
                                        t['signal'],
                                        t['format'],
                                        t['valid'],
                                        t['trigger'],
                                        block=False)
                else:
                    self.sim.trace_buf(t['signal'],
                                       t['size'],
                                       t['valid'],
                                       t['trigger'],
                                       block=False)
        elif cmd == self.ID_MP_ADD_TO_NEW_VIEWER:
            self.AddSelectedToGrid(sim.propgrid())
        else:
            idx = cmd - wx.ID_HIGHEST
            if idx >= 0 and idx < len(grids):
                self.AddSelectedToGrid(grids[idx])

    def AddSelectedToGrid(self, grid):
        if not grid:
            return
        ids = self.tree.GetSelections()
        objs = []
        for item in ids:
            if item == self.tree.GetRootItem():
                continue
            if not item.IsOk():
                break
            ext = self.tree.GetExtendObj(item)
            nkind = ext['nkind']
            self.sim.monitor(ext['name'], grid)
            objs.append(ext['name'])
            if nkind == SC_OBJ_XSC_ARRAY:
                (child, cookie) = self.tree.GetFirstChild(item)
                if child.IsOk() and self.tree.GetItemText(child) == "...":
                    self.tree.Expand(item)
                (child, cookie) = self.tree.GetFirstChild(item)
                while child.IsOk():
                    ext2 = self.tree.GetExtendObj(child)
                    prop = self.sim.monitor(ext2['name'], grid)
                    objs.append(ext2['name'])
                    prop.SetIndent(1)
                    (child, cookie) = self.tree.GetNextChild(item, cookie)
        self.sim.read(objects=objs, block=False)

    def OnTreeBeginDrag(self, event):
        if self.tree.objects is None:
            return

        ids = self.tree.GetSelections()
        objs = []
        objs_name = []
        gname = self.sim.global_object_name
        for item in ids:
            if item == self.tree.GetRootItem():
                continue
            if not item.IsOk():
                break
            ext = self.tree.GetExtendObj(item)
            nkind = ext['nkind']
            if nkind == SC_OBJ_XSC_ARRAY:
                (child, cookie) = self.tree.GetFirstChild(item)
                if child.IsOk() and self.tree.GetItemText(child) == "...":
                    # item has children, but haven't been expanded
                    self.tree.Expand(item)
                (child, cookie) = self.tree.GetFirstChild(item)
                objchild = []
                while child.IsOk():
                    ext2 = self.tree.GetExtendObj(child)
                    objchild.append(gname(ext2['name']))
                    objs_name.append(ext2['name'])
                    (child, cookie) = self.tree.GetNextChild(item, cookie)
                objs.append({'reg': gname(ext['name']), 'child': objchild})
            else:
                objs.append({'reg': gname(ext['name'])})
            objs_name.append(ext['name'])
        # need to explicitly allow drag
        # start drag operation
        data = wx.TextDataObject(json.dumps(objs))
        source = wx.DropSource(self.tree)
        source.SetData(data)
        rtn = source.DoDragDrop(True)
        if rtn == wx.DragError:
            wx.LogError("An error occurred during drag and drop operation")
        elif rtn == wx.DragNone:
            pass
        elif rtn == wx.DragCopy:
            pass
        elif rtn == wx.DragMove:
            pass
        elif rtn == wx.DragCancel:
            pass
        self.sim.read(objects=objs_name, block=False)

    def OnUpdateCmdUI(self, event):
        eid = event.GetId()
        if eid in [self.ID_SIM_STEP, self.ID_SIM_RUN]:
            event.Enable(self.sim and not self.sim.is_running())
        elif eid == self.ID_SIM_PAUSE:
            event.Enable(self.sim and self.sim.is_running())

    def _exit_sim(self):
        msg = 'Do you want to kill %s?' % (self.GetLabel())
        # use top level frame as parent, otherwise it may crash when
        # it is called in Destroy()
        dlg = wx.MessageDialog(self.GetTopLevelParent(), msg, 'bsmedit',
                               wx.YES_NO)
        result = dlg.ShowModal() == wx.ID_YES
        dlg.Destroy()
        if result:
            self.sim.stop()
            wx.CallAfter(dp.send, signal='frame.delete_panel', panel=self)

    def OnProcessCommand(self, event):
        """process the menu command"""
        eid = event.GetId()
        if eid == wx.ID_EXIT:
            # delay destroy the simulation window so AuiToolBar can finish
            # processing the dropdown event.
            wx.CallAfter(self._exit_sim)
        elif eid == wx.ID_RESET:
            self.sim.reset()
        elif eid == self.ID_SIM_STEP:
            self.SetParameter(False)
            self.sim.step()
        elif eid == self.ID_SIM_RUN:
            self.SetParameter(False)
            self.sim.run()
        elif eid == self.ID_SIM_PAUSE:
            self.sim.pause()

    def _process_response(self, resp):
        if self.is_destroying:
            return
        try:
            command = resp.get('cmd', '')
            value = resp.get('value', False)
            if command == 'load':
                self.tree.Load(self.sim.objects)
                dp.send('sim.loaded', num=self.sim.num)
            elif command == 'exit':
                self.tree.Load(None)
                dp.send('sim.unloaded', num=self.sim.num)
                if Gcs.get_active() == self.sim:
                    dp.send(signal="frame.show_status_text", text="")
            elif command == 'monitor_signal':
                objs = [name for name, v in six.iteritems(value) if v]
                self.sim.read(objects=objs, block=False)
            elif command == 'set_parameter':
                if value:
                    step = self.tcStep.GetValue()
                    self.tcStep.SetValue(value.get('step', step))
                    unitStep = self.cmbUnitStep.GetSelection()
                    self.cmbUnitStep.SetSelection(
                        value.get('step_unit', unitStep))
                    total = self.tcTotal.GetValue()
                    self.tcTotal.SetValue(value.get('total', total))
                    unitTotal = self.cmbUnitTotal.GetSelection()
                    self.cmbUnitTotal.SetSelection(
                        value.get('total_unit', unitTotal))
            elif command == 'read':
                gname = self.sim.global_object_name
                ui_objs = {gname(name): v for name, v in six.iteritems(value)}
                dp.send(signal="sim.update", objs=ui_objs)
            elif command == 'read_buf':
                gname = self.sim.global_object_name
                ui_buffers = {
                    gname(name): v
                    for name, v in six.iteritems(value)
                }
                dp.send(signal="sim.buffer_changed", bufs=ui_buffers)
            elif command == 'time_stamp':
                if isinstance(value, six.string_types):
                    dp.send(signal="frame.show_status_text", text=value)
            elif command == 'breakpoint_triggered':
                SimPropGrid.ClearAllTrigger()
                bp = value  #[name, condition, hitcount, hitsofar]
                gname = self.sim.global_object_name
                for grid in SimPropGrid.GCM.get_all_managers():
                    if grid.TriggerBreakPoint(gname(bp[0]), bp[1], bp[2]):
                        dp.send(signal='frame.show_panel', panel=grid)
                        break
                else:
                    txt = "Breakpoint triggered: %s\n" % (json.dumps(bp))
                    dp.send(signal='shell.write_out', text=txt)
            elif command == 'write_out':
                dp.send(signal='shell.write_out', text=value)

            if command in ['load', 'step'] and value:
                self.sim.time_stamp(insecond=False, block=False)
                self.sim.read(objects=[], block=False)
                self.sim.read_buf(objects=[], block=False)
        except:
            traceback.print_exc(file=sys.stdout)


wxEVT_PROP_CLICK_CHECK = wx.NewEventType()
EVT_PROP_CLICK_CHECK = wx.PyEventBinder(wxEVT_PROP_CLICK_CHECK, 1)


class SimProperty(pg.Property):
    def __init__(self, grid, name, label, value):
        pg.Property.__init__(self, grid, name, label, value)
        self.gripper_clr = wx.RED
        self.show_check = True
        self.checked = False
        self.condition = ("", "")
        self.triggered = False

    def duplicate(self):
        p = super(SimProperty, self).duplicate()
        p.gripper_clr = self.gripper_clr
        p.show_check = self.show_check
        p.checked = self.checked
        return p

    def SetGripperColor(self, clr=None):
        self.gripper_clr = clr

    def GetGripperColor(self):
        return self.gripper_clr

    def SetShowCheck(self, show=True, silent=True):
        """show/hide radio button"""
        if self.show_check == show:
            return
        self.show_check = show
        if not silent:
            self.Refresh()

    def IsShowCheck(self):
        """return whether the icon is shown"""
        return self.show_check

    def SetChecked(self, check=True, silent=False):
        """check/uncheck the radio button"""
        if check != self.IsChecked():
            self.checked = check
            if not self.SendPropEvent(wxEVT_PROP_CLICK_CHECK):
                self.checked = not check
            if not silent:
                self.Refresh()

    def IsChecked(self):
        """return true if the radio button is checked"""
        return self.checked

    def OnMouseUp(self, pt):
        ht = self.HitTest(pt)
        if self.IsEnabled():
            # click on the check icon? change the state
            if self.IsShowCheck() and ht == 'check':
                checked = self.IsChecked()
                self.SetChecked(not checked)
        return ht

    def SetBpConditon(self, cond, hitcount):
        self.condition = (cond, hitcount)

    def GetBpCondition(self):
        return self.condition

    def SetTriggered(self, triggered):
        self.triggered = triggered

    def IsTriggered(self):
        return self.triggered

class SimPropArt(pg.PropArtNative):
    def __init__(self):
        super(SimPropArt, self).__init__()
        self.gripper_width = 6
        if wx.Platform == '__WXMSW__':
            self.img_expand = wx.ImageList(12, 12, True, 2)
            self.img_expand.Add(wx.Bitmap(to_byte(pg.tree_xpm)))
            self.expansion_width = 12
        self.check_width = 16
        sx, sy = 16, 16
        self.img_check = wx.ImageList(sx, sy, True, 4)
        self.img_check.Add(svg_to_bitmap(radio_unchecked_svg, width=sx, height=sy))
        self.img_check.Add(svg_to_bitmap(radio_disabled_svg, width=sx, height=sy))
        self.img_check.Add(svg_to_bitmap(radio_checked_svg, width=sx, height=sy))
        self.img_check.Add(svg_to_bitmap(radio_activated_svg, width=sx, height=sy))
        #self.img_check.Add(wx.Bitmap(to_byte(pg.radio_xpm)))

    def PrepareDrawRect(self, p):
        """calculate the rect for each section"""
        mx = self.gap_x
        irc = p.GetRect()
        irc.SetLeft(irc.GetLeft() + self.margin['left'])
        irc.SetRight(irc.GetRight() + self.margin['right'])
        irc.SetTop(irc.GetTop() + self.margin['top'])
        irc.SetBottom(irc.GetBottom() + self.margin['bottom'])
        x = irc.x
        x = x + mx * 2 + p.indent * self.indent_width
        # gripper
        rc = wx.Rect(*irc)
        rc.x += self.gap_x
        rc.SetWidth(self.gripper_width)
        p.regions['gripper'] = rc

        if self.expansion_width > 0 and p.HasChildren():
            # expander icon
            rc = wx.Rect(*irc)
            rc.x = x + mx * 2
            rc.SetWidth(self.expansion_width)
            p.regions['expander'] = rc
            x = rc.right

        if self.check_width > 0 and p.IsShowCheck():
            # radio/check icon
            rc = wx.Rect(*irc)
            rc.x = x + mx
            rc.SetWidth(self.check_width + 2)
            p.regions['check'] = rc
            x = rc.right

        # label
        p.regions['label'] = wx.Rect(*irc)
        p.regions['label'].x = x + mx * 2

        if not p.IsSeparator():
            title_width = p.title_width
            if title_width < 0:
                title_width = self.title_width
            p.regions['label'].SetRight(title_width)
            x = p.regions['label'].right

            rc = wx.Rect(*irc)
            rc.x = x + mx
            rc.SetWidth(self.splitter_width)
            p.regions['splitter'] = rc
            x = rc.right

            rc = wx.Rect(*irc)
            rc.x = x
            rc.SetWidth(irc.right - x)
            p.regions['value'] = rc
        else:
            # separator does not have splitter & value
            p.regions['label'].SetWidth(p.regions['label'].GetWidth() +
                                        irc.right - x)
            p.regions['splitter'] = wx.Rect(irc.right, irc.top, 0, 0)
            p.regions['value'] = wx.Rect(irc.right, irc.top, 0, 0)

    def DrawGripper(self, dc, p):
        # draw gripper
        if p.gripper_clr:
            pen = wx.Pen(wx.BLACK, 1, wx.PENSTYLE_TRANSPARENT)
            dc.SetPen(pen)

            dc.SetBrush(wx.Brush(p.gripper_clr))
            rc = p.regions['gripper']
            dc.DrawRectangle(rc.x, rc.y + 1, 3, rc.height - 1)

    def DrawCheckNative(self, dc, p):
        # draw radio button
        if self.check_width > 0 and p.IsShowCheck():
            render = wx.RendererNative.Get()
            state = 0
            if not p.IsEnabled():
                state |= wx.CONTROL_DISABLED
            if p.IsChecked():
                state |= wx.CONTROL_CHECKED
            if p.IsActivated():
                state |= wx.CONTROL_FOCUSED

            w, h = self.check_width, self.check_width
            rc = p.regions['check']
            x = rc.x + (rc.width - w) / 2
            y = rc.y + (rc.height - h) / 2 + 1
            render.DrawRadioBitmap(p.grid, dc, (x, y, w, h), state)

    def DrawCheck(self, dc, p):
        if self.check_width > 0 and p.IsShowCheck():
            state = 0
            if not p.IsEnabled():
                state = 1
            elif p.IsChecked():
                state = 2
                if p.IsTriggered():
                    state = 3

            if self.img_check.GetImageCount() == 4:
                (w, h) = self.img_check.GetSize(0)
                rc = p.regions['check']
                x = rc.x + (rc.width - w) / 2
                y = rc.y + (rc.height - h) / 2 + 1
                self.img_check.Draw(state, dc, x, y,
                                    wx.IMAGELIST_DRAW_TRANSPARENT)
            else:
                self.DrawCheckNative(dc, p)

    def DrawExpansion(self, dc, p):
        if p.HasChildren():
            if hasattr(self, 'img_expand') and self.img_expand.GetImageCount() == 2:
                (w, h) = self.img_expand.GetSize(0)
                rc = p.regions['expander']
                x = rc.x + (rc.width - w) / 2
                y = rc.y + (rc.height - h) / 2 + 1
                idx = 0
                if not p.IsExpanded():
                    idx = 1
                self.img_expand.Draw(idx, dc, x, y,
                                     wx.IMAGELIST_DRAW_TRANSPARENT)
            else:
                super(SimPropArt, self).DrawExpansion(dc, p)

    def DrawItem(self, dc, p):
        super(SimPropArt, self).DrawItem(dc, p)
        self.DrawGripper(dc, p)
        self.DrawCheck(dc, p)


class SimPropGrid(pg.PropGrid):
    GCM = Gcm()
    ID_PROP_BREAKPOINT = wx.NewId()
    ID_PROP_BREAKPOINT_CLEAR = wx.NewId()

    def __init__(self, parent, num=None):
        pg.PropGrid.__init__(self, parent)

        # if num is not defined or is occupied, generate a new one
        if num is None or num in SimPropGrid.GCM.get_nums():
            num = SimPropGrid.GCM.get_next_num()
        self.num = num
        SimPropGrid.GCM.set_active(self)
        self.SetArtProvider(SimPropArt())
        self._prop_cls = SimProperty
        self.Bind(EVT_PROP_CLICK_CHECK, self.OnPropEventsHandler)

        dp.connect(self.OnSimLoad, 'sim.loaded')
        dp.connect(self.OnSimUnload, 'sim.unloaded')
        dp.connect(self.OnSimUpdate, 'sim.update')

    def Destroy(self):
        dp.disconnect(self.OnSimLoad, 'sim.loaded')
        dp.disconnect(self.OnSimUnload, 'sim.unloaded')
        dp.disconnect(self.OnSimUpdate, 'sim.update')
        self.DeleteAllProperties()
        SimPropGrid.GCM.destroy_mgr(self)
        super(SimPropGrid, self).Destroy()

    def OnSimUpdate(self, objs):
        for name, v in six.iteritems(objs):
            p = self.GetProperty(name)
            if isinstance(p, list):
                for prop in p:
                    prop.SetValue(v)
                    prop.SetValueValid(True)
            elif isinstance(p, pg.Property):
                p.SetValue(v)
                p.SetValueValid(True)

    def OnSimLoad(self, num):
        """try reconnecting registers when the simulation is loaded."""
        s = str(num) + '.'
        sl = len(s)
        objs = []
        for p in self._props:
            name = p.GetName()
            if name.startswith(s):
                objs.append(name[sl:])
        if objs:
            # tell the simulation to monitor signals
            resp = dp.send('sim.command',
                           num=num,
                           command='monitor_signal',
                           objects=objs)
            if not resp:
                return
            status = resp[0][1]
            if status == False:
                return
            for obj in objs:
                # enable props that is monitored successfully
                if isinstance(status, dict) and not status.get(obj, False):
                    continue
                p = self.GetProperty(s + obj)
                if not p:
                    continue
                if isinstance(p, pg.Property):
                    p = [p]
                for prop in p:
                    prop.SetReadonly(False)
                    prop.Enable(True)

    def OnSimUnload(self, num):
        # disable all props from the simulation with id 'num'
        s = str(num) + '.'
        for p in self._props:
            name = p.GetName()
            if not name.startswith(s):
                continue
            p.SetReadonly(True)
            p.Enable(False)

    def GetContextMenu(self, prop):
        menu = super(SimPropGrid, self).GetContextMenu(prop)
        if not menu:
            menu = wx.Menu()
        else:
            menu.AppendSeparator()
        if prop:
            menu.Append(self.ID_PROP_BREAKPOINT, "Breakpoint Condition")
            menu.Enable(self.ID_PROP_BREAKPOINT, prop.IsChecked())
        menu.Append(self.ID_PROP_BREAKPOINT_CLEAR, "Clear all Breakpoints")
        return menu

    def OnPropEventsHandler(self, evt):
        if not super(SimPropGrid, self).OnPropEventsHandler(evt):
            return False

        prop = evt.GetProperty()
        eid = evt.GetEventType()
        if eid == wxEVT_PROP_CLICK_CHECK:
            # turn on/off breakpoint
            if prop.IsChecked():
                dp.send('prop.bp_add', prop=prop)
            else:
                dp.send('prop.bp_del', prop=prop)
        elif eid == pg.wxEVT_PROP_CHANGED:
            # the value changed, notify the parent
            dp.send('prop.changed', prop=prop)
        return True

    def OnProcessCommand(self, eid, prop):
        """process the context menu command"""
        if eid == self.ID_PROP_BREAKPOINT:
            if not prop:
                return
            condition = prop.GetBpCondition()
            dlg = BreakpointSettingsDlg(self, condition[0], condition[1])
            if dlg.ShowModal() == wx.ID_OK:
                if prop.IsChecked():
                    # delete current breakpoint
                    prop.SetChecked(False)
                    prop.SetBpConditon(*dlg.GetCondition())
                    # add breakpoint to make the condition valid
                    prop.SetChecked(True)
                else:
                    prop.SetBpConditon(*dlg.GetCondition())

        elif eid == self.ID_PROP_BREAKPOINT_CLEAR:
            self.ClearBreakPoints()
        else:
            super(SimPropGrid, self).OnProcessCommand(eid, prop)

    def ClearBreakPoints(self):
        """clear all the breakpoints"""
        for prop in self._props:
            if prop and prop.IsChecked():
                prop.SetChecked(False)
    @classmethod
    def ClearAllTrigger(cls):
        for grid in SimPropGrid.GCM.get_all_managers():
            grid.ClearTrigger()

    def ClearTrigger(self):
        for prop in self._props:
            prop.SetTriggered(False)

    def TriggerBreakPoint(self, name, cond, hitcount):
        """check whether the breakpoints are triggered"""
        for prop in self._props:
            if name == prop.GetName():
                if (cond, hitcount) == prop.GetBpCondition():
                    prop.SetTriggered(True)
                    self.EnsureVisible(prop)
                    self.SetSelection(prop)
                    return True
        return False


class BreakpointSettingsDlg(wx.Dialog):
    def __init__(self, parent, condition='', hitcount=''):
        wx.Dialog.__init__(self,
                           parent,
                           title="Breakpoint Condition",
                           size=wx.Size(431, 289),
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        szAll = wx.BoxSizer(wx.VERTICAL)
        label = ("At the end of each delta cycle, the expression is evaluated "
                 "and the breakpoint is hit only if the expression is true or "
                 "the register value has changed")
        self.stInfo = wx.StaticText(self, label=label)
        self.stInfo.Wrap(400)
        szAll.Add(self.stInfo, 1, wx.ALL, 5)

        szCnd = wx.BoxSizer(wx.HORIZONTAL)

        szCond = wx.BoxSizer(wx.VERTICAL)

        self.rbChanged = wx.RadioButton(self,
                                        label="Has changed",
                                        style=wx.RB_GROUP)
        szCond.Add(self.rbChanged, 5, wx.ALL | wx.EXPAND, 5)

        label = "Is true (value: $; for example, $==10)"
        self.rbCond = wx.RadioButton(self, label=label)
        szCond.Add(self.rbCond, 0, wx.ALL | wx.EXPAND, 5)

        self.tcCond = wx.TextCtrl(self)
        szCond.Add(self.tcCond, 0, wx.ALL | wx.EXPAND, 5)

        label = "Hit count (hit count: #; for example, #>10"
        self.cbHitCount = wx.CheckBox(self, label=label)
        szCond.Add(self.cbHitCount, 0, wx.ALL, 5)

        self.tcHitCount = wx.TextCtrl(self)
        szCond.Add(self.tcHitCount, 0, wx.ALL | wx.EXPAND, 5)

        szCnd.Add(szCond, 1, wx.EXPAND, 5)

        szAll.Add(szCnd, 1, wx.EXPAND, 5)

        self.stLine = wx.StaticLine(self, style=wx.LI_HORIZONTAL)
        szAll.Add(self.stLine, 0, wx.EXPAND | wx.ALL, 5)

        szConfirm = wx.BoxSizer(wx.HORIZONTAL)

        self.btnOK = wx.Button(self, wx.ID_OK, u"OK")
        szConfirm.Add(self.btnOK, 0, wx.ALL, 5)

        self.btnCancel = wx.Button(self, wx.ID_CANCEL, u"Cancel")
        szConfirm.Add(self.btnCancel, 0, wx.ALL, 5)

        szAll.Add(szConfirm, 0, wx.ALIGN_RIGHT, 5)

        self.SetSizer(szAll)
        self.Layout()

        # initialize the controls
        self.condition = condition
        self.hitcount = hitcount
        if self.condition == '':
            self.rbChanged.SetValue(True)
            self.tcCond.Disable()
        else:
            self.rbChanged.SetValue(False)
            self.rbCond.SetValue(True)
        self.tcCond.SetValue(self.condition)
        if self.hitcount == '':
            self.cbHitCount.SetValue(False)
            self.tcHitCount.Disable()
        else:
            self.cbHitCount.SetValue(True)
        self.tcHitCount.SetValue(self.hitcount)
        # Connect Events
        self.rbChanged.Bind(wx.EVT_RADIOBUTTON, self.OnRadioButton)
        self.rbCond.Bind(wx.EVT_RADIOBUTTON, self.OnRadioButton)
        self.cbHitCount.Bind(wx.EVT_CHECKBOX, self.OnRadioButton)
        self.btnOK.Bind(wx.EVT_BUTTON, self.OnBtnOK)

    def OnRadioButton(self, event):
        if self.rbChanged.GetValue():
            self.tcCond.Disable()
        else:
            self.tcCond.Enable()
        if self.cbHitCount.GetValue():
            self.tcHitCount.Enable()
        else:
            self.tcHitCount.Disable()
        event.Skip()

    def OnBtnOK(self, event):
        # set condition to empty string to indicate the breakpoint will be
        # triggered when the value is changed
        if self.rbChanged.GetValue():
            self.condition = ''
        else:
            self.condition = self.tcCond.GetValue()
        if self.cbHitCount.GetValue():
            self.hitcount = self.tcHitCount.GetValue()
        else:
            self.hitcount = ""
        event.Skip()

    def GetCondition(self):
        return (self.condition, self.hitcount)


class sim(object):
    frame = None
    ID_SIM_NEW = wx.NOT_FOUND
    ID_PROP_NEW = wx.NOT_FOUND

    @classmethod
    def initialize(cls, frame):
        cls.frame = frame

        resp = dp.send(signal='frame.add_menu',
                       path='File:New:Simulation',
                       rxsignal='bsm.simulation')
        if resp:
            cls.ID_SIM_NEW = resp[0][1]
        resp = dp.send(signal='frame.add_menu',
                       path='File:New:PropGrid',
                       rxsignal='bsm.simulation')
        if resp:
            cls.ID_PROP_NEW = resp[0][1]

        dp.connect(cls._process_command, signal='bsm.simulation')
        dp.connect(cls._sim_command, signal='sim.command')
        dp.connect(receiver=cls._frame_set_active,
                   signal='frame.activate_panel')
        dp.connect(receiver=cls._frame_uninitialize, signal='frame.exit')
        dp.connect(receiver=cls.initialized, signal='frame.initialized')
        dp.connect(receiver=cls._prop_insert, signal='prop.insert')
        dp.connect(receiver=cls._prop_delete, signal='prop.delete')
        dp.connect(receiver=cls._prop_drop, signal='prop.drop')
        dp.connect(receiver=cls._prop_bp_add, signal='prop.bp_add')
        dp.connect(receiver=cls._prop_bp_del, signal='prop.bp_del')
        dp.connect(receiver=cls._prop_changed, signal='prop.changed')

    @classmethod
    def initialized(cls):
        dp.send(signal='shell.run',
                command='from bsmedit.bsm.pysim import *',
                prompt=False,
                verbose=False,
                history=False)
        dp.send(signal='shell.run',
                command='import six',
                prompt=False,
                verbose=False,
                history=False)

    @classmethod
    def _sim_command(cls, num, command, **kwargs):
        mgr = Gcs.get_manager(num)
        kwargs.pop("signal", None)
        kwargs.pop("sender", None)
        if mgr and hasattr(mgr, command) and command in mgr._interfaces:
            fun = getattr(mgr, command)
            if fun:
                return fun(**kwargs)
        return False

    @classmethod
    def _prop_changed(cls, prop):
        mgr, name = cls._find_object(prop.GetName())
        if mgr:
            mgr.write({name: prop.GetValue()})

    @classmethod
    def _prop_bp_add(cls, prop):
        mgr, name = cls._find_object(prop.GetName())
        if mgr:
            cnd = prop.GetBpCondition()
            if cnd is None:
                cnd = ["", ""]
            mgr.add_breakpoint(name, cnd[0], cnd[1])

    @classmethod
    def _prop_bp_del(cls, prop):
        mgr, name = cls._find_object(prop.GetName())
        if mgr:
            cnd = prop.GetBpCondition()
            if cnd is None:
                cnd = ("", "")
            mgr.del_breakpoint(name, cnd[0], cnd[1])

    @classmethod
    def _prop_insert(cls, prop):
        mgr, name = cls._find_object(prop.GetName())
        if mgr:
            mgr.monitor_signal(name)

    @classmethod
    def _prop_delete(cls, prop):
        mgr, name = cls._find_object(prop.GetName())
        if mgr:
            mgr.unmonitor_signal(name)

    @classmethod
    def _prop_drop(cls, index, prop, grid):
        objs = json.loads(prop)
        for obj in objs:
            reg = obj['reg']
            mgr, name = cls._find_object(str(reg))
            if not mgr:
                continue
            p = mgr.monitor(name, grid, index)
            if index != -1:
                index = index + 1
            for c in obj.get('child', []):
                mgr, name = cls._find_object(str(c))
                if not mgr:
                    continue
                p = mgr.monitor(name, grid, index)
                p.SetIndent(1)
                if index != -1:
                    index = index + 1

    @classmethod
    def _frame_set_active(cls, pane):
        if pane and isinstance(pane, SimPanel):
            if Gcs.get_active() == pane.sim:
                return
            Gcs.set_active(pane.sim)
            if pane.sim.is_valid() and not pane.sim.is_running():
                # update the time stamp on the status bar
                pane.sim.time_stamp(insecond=False, block=False)
        if pane and isinstance(pane, SimPropGrid):
            SimPropGrid.GCM.set_active(pane)

    @classmethod
    def _frame_uninitialize(cls):
        for mgr in Gcs.get_all_managers():
            mgr.stop()
            dp.send('frame.delete_panel', panel=mgr.frame)
        dp.send('frame.delete_menu', path="View:Simulations")
        dp.send('frame.delete_menu',
                path="File:New:Simulation",
                id=cls.ID_SIM_NEW)
        dp.send('frame.delete_menu',
                path="File:New:PropGrid",
                id=cls.ID_PROP_NEW)

    @classmethod
    def _process_command(cls, command):
        if command == cls.ID_SIM_NEW:
            style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
            dlg = wx.FileDialog(cls.frame, "Choose a file", "", "", "*.*",
                                style)
            if dlg.ShowModal() == wx.ID_OK:
                filename = dlg.GetPath()
                cls.simulation(filename=filename)
            dlg.Destroy()
        if command == cls.ID_PROP_NEW:
            cls.propgrid()

    @classmethod
    def _find_object(cls, name):
        """
        Find the object in simulations by its name, for example
            '1.top.sig_cos' -> find the object with name 'top.sig_cos' in
                               simulation 1
            'top.sig_cos' -> find the object with name 'top.sig_cos' in
                             the active simulation, the active simulation is
                             the simulation that is currently active or most
                             recently active.
        """
        x = re.match(r'^(\d)+\.(.*)', name)
        if x:
            num, obj = (int(x.group(1)), x.group(2))
        else:
            num, obj = None, name
        if num is None:
            mgr = Gcs.get_active()
        else:
            mgr = Gcs.get_manager(num)
        if not mgr or obj not in mgr.objects:
            return None, None
        return mgr, obj

    @classmethod
    def simulation(cls,
                   num=None,
                   filename=None,
                   silent=False,
                   create=True,
                   activate=False):
        """
        create a simulation

        If the simulation exists, return its handler; otherwise, create it if
        create == True.
        """
        def GetColorByNum(num):
            color = [
                'green', 'red', 'blue', 'black', 'cyan', 'yellow', 'magenta',
                'cyan'
            ]
            return wx.Colour(color[num % len(color)])

        manager = Gcs.get_manager(num)
        if manager is None and create:
            manager = SimPanel(sim.frame, num, filename, silent)
            clr = GetColorByNum(manager.sim.num)
            clr.Set(clr.red, clr.green, clr.blue, 128)
            manager.SetColor(clr)
            page_bmp = MakeBitmap(clr.red, clr.green,
                                  clr.blue)  #178,  34,  34)
            title = "Simulation-%d" % manager.sim.num
            dp.send(signal="frame.add_panel",
                    panel=manager,
                    title=title,
                    target="History",
                    icon=page_bmp,
                    showhidemenu="View:Simulations:%s" % title)
            return manager.sim
        # activate the manager
        elif manager and activate:
            dp.send(signal='frame.show_panel', panel=manager)

        return manager

    @classmethod
    def propgrid(cls, num=None, create=True, activate=False):
        """
        get the propgrid handle by its number

        If the propgrid exists, return its handler; otherwise, it will be created.
        """
        mgr = SimPropGrid.GCM.get_manager(num)
        if not mgr and create:
            mgr = SimPropGrid(cls.frame)
            mgr.SetLabel("Propgrid-%d" % mgr.num)
            dp.send(signal="frame.add_panel", panel=mgr, title=mgr.GetLabel())
        elif mgr and activate:
            # activate the window
            dp.send(signal='frame.show_panel', panel=mgr)
        return mgr

    @classmethod
    def plot_trace(cls, *args, **kwargs):
        """
        plot the trace

        The trace will be automatically updated after each simulation step.
        """
        if args and isinstance(args[0], six.string_types):
            s1, obj1 = cls._find_object(args[0])
            if not s1 or not obj1:
                print("unknown sim object:", args[0])
                return
            s2, obj2 = None, None
            if len(args) > 1 and isinstance(args[1], six.string_types):
                s2, obj2 = cls._find_object(args[1])

        if obj2:
            args = args[2:]
            sx, x, sy, y = s1, obj1, s2, obj2
        else:
            args = args[1:]
            sx, x, sy, y = None, None, s1, obj1

        dy = sy.read_buf(y, block=True)
        y = {sy.global_object_name(y): dy}
        if sx:
            dx = sx.read_buf(x, block=True)
            x = {sx.global_object_name(x): dx}
        # since matplotlib 3.1.0, get_current_fig_manager returns an instance
        # of FigureManagerBase, and its frame is our panel
        mgr = graph.plt.get_current_fig_manager()
        if not isinstance(mgr, graph.MatplotPanel) and hasattr(mgr, 'frame'):
            mgr = mgr.frame
        autorelim = kwargs.pop("relim", True)
        mgr.plot_trace(x, y, autorelim, *args, **kwargs)


def bsm_initialize(frame, **kwargs):
    sim.initialize(frame)



import sys
import traceback
import json
import six
import wx
import wx.py.dispatcher as dp
import wx.lib.agw.aui as aui
from .prop import *
from .formatters import *
from .propart import PropArtNative

wxEVT_PROP_INSERT = wx.NewEventType()
wxEVT_PROP_DELETE = wx.NewEventType()

EVT_PROP_INSERT = wx.PyEventBinder(wxEVT_PROP_INSERT, 1)
EVT_PROP_DELETE = wx.PyEventBinder(wxEVT_PROP_DELETE, 1)


def PopupMenu(wnd, menu):
    if not wnd or not menu:
        return wx.ID_NONE

    cc = aui.ToolbarCommandCapture()
    wnd.PushEventHandler(cc)

    wnd.PopupMenu(menu)

    command = cc.GetCommandId()
    wnd.PopEventHandler(True)
    return command


class PropDropTarget(wx.DropTarget):
    def __init__(self, frame):
        wx.DropTarget.__init__(self)
        self.obj = wx.TextDataObject()
        self.SetDataObject(self.obj)
        self.frame = frame

    def OnEnter(self, x, y, d):
        return super(PropDropTarget, self).OnDragOver(x, y, d)

    def OnLeave(self):
        pass

    def OnData(self, x, y, d):
        if not self.GetData():
            return wx.DragNone
        self.frame.OnDrop(x, y, self.obj.GetText())

        return d

    def OnDragOver(self, x, y, d):
        pt = wx.Point(x, y)
        rc = self.frame.GetClientRect()
        if rc.Contains(pt):
            (x, y) = self.frame.GetViewStart()
            if pt.y < 15:
                self.frame.Scroll(-1, y - (15 - pt.y) / 3)
            if pt.y > rc.bottom - 15:
                self.frame.Scroll(-1, y - (rc.bottom - 15 - pt.y) / 3)
        return super(PropDropTarget, self).OnDragOver(x, y, d)


class PropGrid(wx.ScrolledWindow):
    ID_PROP_GRID_ADD_SEP = wx.NewId()
    ID_PROP_GRID_READ_ONLY = wx.NewId()
    ID_PROP_GRID_INDENT_INS = wx.NewId()
    ID_PROP_GRID_INDENT_DES = wx.NewId()
    ID_PROP_GRID_MOVE_UP = wx.NewId()
    ID_PROP_GRID_MOVE_DOWN = wx.NewId()
    ID_PROP_GRID_DELETE = wx.NewId()
    ID_PROP_GRID_PROP = wx.NewId()

    drag_state = 0
    drag_start = wx.Point(0, 0)
    drag_prop = None
    drag_pg = None

    RESIZE_NONE = 0
    RESIZE_SEP = 1
    RESIZE_BOT = 2

    SCROLL_UNIT = 15

    CURSOR_RESIZE_HORZ = 0
    CURSOR_RESIZE_VERT = 1
    CURSOR_STD = 2

    def __init__(self, frame, prop_cls=Property):
        wx.ScrolledWindow.__init__(self, frame)

        self._prop_cls = prop_cls
        self.prop_selected = None
        self.cursor_mode = self.CURSOR_STD
        self.pos_mouse_down = wx.Point(0, 0)
        self.prop_under_mouse = None
        self.resize_mode = self.RESIZE_NONE

        self._props = []
        self._art = PropArtNative()

        # cursor
        self.resize_cursor_horz = wx.Cursor(wx.CURSOR_SIZEWE)
        self.resize_cursor_vert = wx.Cursor(wx.CURSOR_SIZENS)

        # set scroll parameters
        self.SetScrollRate(self.SCROLL_UNIT, self.SCROLL_UNIT)
        self.SetVirtualSize(wx.Size(100, 200))

        self.SetDropTarget(PropDropTarget(self))

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnMouseDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnMouseUp)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnMouseRightClick)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnMouseDoubleClick)
        self.Bind(wx.EVT_MOTION, self.OnMouseMove)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnMouseLeave)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        # calling CaptureMouse requires to implement EVT_MOUSE_CAPTURE_LOST
        self.Bind(wx.EVT_MOUSE_CAPTURE_LOST, self.OnMouseCaptureLost)

        self.Bind(EVT_PROP_SELECTED, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_CHANGING, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_CHANGED, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_HIGHLIGHTED, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_RIGHT_CLICK, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_COLLAPSED, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_EXPANDED, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_DOUBLE_CLICK, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_INDENT, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_KEYDOWN, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_RESIZE, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_REFRESH, self.OnPropRefresh)
        self.Bind(EVT_PROP_DROP, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_BEGIN_DRAG, self.OnPropEventsHandler)
        self.Bind(wx.EVT_MENU, self.OnProcessCommand)
        self.SetBackgroundStyle(wx.BG_STYLE_PAINT)

    def SetArtProvider(self, art):
        self._art = art
        self.Refresh()

    def GetArtProvider(self):
        return self._art

    def AppendProperty(self, name, label="", value="", update=True):
        return self.InsertProperty(name, label, value, -1, update)

    def _InsertProperty(self, prop, index=-1, update=True):
        # add the prop window to the grid
        if not isinstance(prop, Property):
            return None

        if index == -1 or index >= self.GetPropCount():
            self._props.append(prop)
        else:
            self._props.insert(index, prop)
        name = prop.GetName()

        if index != -1 and (not update):
            self.CheckProp()
        if update:
            self.UpdateGrid()
        if self.SendPropEvent(wxEVT_PROP_INSERT, prop):
            dp.send('prop.insert', prop=prop)
        return prop

    def InsertProperty(self, name, label="", value="", index=-1, update=True):
        # add the prop window to the grid
        prop = self._prop_cls(self, name, label, value)
        return self._InsertProperty(prop, index, update)

    def CopyProperty(self, prop, index=-1, update=True):
        if not isinstance(prop, Property):
            return None
        p = prop.duplicate()
        p.SetParent(self)
        return self._InsertProperty(p, index, update)

    def InsertSeparator(self,
                        name='Separator',
                        label='',
                        index=-1,
                        update=True):
        prop = self.InsertProperty(name, label, '', index, update)
        if prop:
            prop.SetSeparator(True)
        return prop

    def RemoveProperty(self, prop, update=True):
        # remove property
        if isinstance(prop, six.string_types) or isinstance(prop, Property):
            index = self.FindPropertyIndex(prop)
        elif isinstance(prop, int):
            index = prop
        else:
            return False

        if index >= 0 and index < self.GetPropCount():
            prop = self._props[index]
            activated = False
            if prop == self.prop_selected:
                activated = True
                self.SetSelection(-1)
            del self._props[index]

            if index != -1 and (not update):
                self.CheckProp()
            if index >= self.GetPropCount():
                index = self.GetPropCount() - 1

            if activated:
                self.SetSelection(index)
            if update:
                self.UpdateGrid()
            return True
        return False

    def DeleteAllProperties(self, update=True):
        for i in range(len(self._props) - 1, -1, -1):
            self.DeleteProperty(self._props[i], update)

    def DeleteProperty(self, prop, update=True):
        if self.SendPropEvent(wxEVT_PROP_DELETE, prop):
            dp.send('prop.delete', prop=prop)
            return self.RemoveProperty(prop, update)
        else:
            return False

    def FindPropertyIndex(self, prop):
        """return the index of prop, or -1 if not found"""
        p = self.GetProperty(prop)
        if not p:
            return -1
        try:
            idx = self._props.index(p)
            return idx
        except ValueError:
            traceback.print_exc(file=sys.stdout)
        return -1

    def GetProperty(self, prop):
        """return the Property instance"""
        if isinstance(prop, Property):
            # if prop is a Property instance, simply return
            return prop
        elif isinstance(prop, six.string_types):
            # search the prop name
            props = [p for p in self._props if p.GetName() == prop]
            if not props:
                return None
            elif len(props) == 1:
                return props[0]
            return props
        elif isinstance(prop, int):
            # prop is the index
            index = prop
            if index >= 0 and index < self.GetPropCount():
                return self._props[index]
        return None

    def GetPropCount(self):
        """return the number of properties"""
        return len(self._props)

    def GetScrolledRect(self, rc):
        (x, y) = self.GetViewStart()
        rcs = wx.Rect(*rc)
        rcs.Offset(wx.Point(-x * self.SCROLL_UNIT, -y * self.SCROLL_UNIT))
        return rcs

    def EnsureVisible(self, prop):
        """scroll the window to make sure prop is visible"""
        p = self.GetProperty(prop)
        if not p:
            return
        rc_prop = p.GetRect()
        # translate to the scrolled position
        rc_prop.x, rc_prop.y = self.CalcScrolledPosition(rc_prop.x, rc_prop.y)
        _, y = self.GetViewStart()
        rc = self.GetRect()
        if rc.top < rc_prop.top and rc.bottom > rc_prop.bottom:
            # if the prop is visible, simply return
            return
        if rc.top > rc_prop.top:
            # if the prop is on top of the client window, scroll up
            y = y + ((rc_prop.top - rc.top) / self.SCROLL_UNIT)
            self.Scroll(-1, y)
        elif rc.bottom < rc_prop.bottom:
            # if the prop is under bottom of the client window, scroll down
            y = y + ((rc_prop.bottom - rc.bottom) / self.SCROLL_UNIT)
            self.Scroll(-1, y)

    def GetSelection(self):
        """get the index of the selected property"""
        return self.FindPropertyIndex(self.prop_selected)

    def GetSelectedProperty(self):
        """return the selected property"""
        return self.prop_selected

    def SetSelection(self, prop):
        """set the active property"""
        p = self.GetProperty(prop)
        if p != self.prop_selected:
            if self.prop_selected:
                self.prop_selected.SetActivated(False)
            self.prop_selected = p
            if self.prop_selected:
                self.prop_selected.SetActivated(True)
            self.Refresh()
            return True
        return False

    def UpdateGrid(self):
        """update the grid"""
        self.LayoutAll()
        self.Refresh()

    def MoveProperty(self, prop, step):
        """move the property"""
        index = self.FindPropertyIndex(prop)
        if index == -1:
            return

        if step == 0:
            # move zero step is no move at all
            return

        # calculate the new position
        index2 = index + step
        if index2 < 0:
            index2 = 0
        # move the prop, prop will be placed on top of index2
        if index2 < self.GetPropCount():
            self.doMoveProperty(index, index2)
        else:
            self.doMoveProperty(index, -1)

    def doMoveProperty(self, index, index2):
        """move the property"""
        # the same position, ignore it
        if index == index2:
            return

        prop = self.GetProperty(index)
        props = [prop]
        if prop.HasChildren() and (not prop.IsExpanded()):
            # move all the children if they are not visible
            indent = prop.GetIndent()
            for i in six.moves.range(index + 1, self.GetPropCount()):
                if self._props[i].GetIndent() <= indent:
                    break
                props.append(self._props[i])

        i = 0
        for p in props:
            if index2 == -1:
                self._props.append(p)
            else:
                #insert it before index2
                self._props.insert(index2 + i, p)
                i += 1

        if index2 != -1 and index > index2:
            index = index + len(props)

        # delete the original properties
        for i in six.moves.range(0, len(props)):
            del self._props[index]

        self.UpdateGrid()

    def MovePropertyDown(self, prop):
        """move the property one step down"""
        # here step is 2 instead of 1 because the prop will be moved in front
        # of index + step. For example, prop is at position 5, to move it to
        # position 6:
        #    step 1) copy it in front of position 7 (position 7);
        #    step 2) remove the original prop at position 5
        #    step 3) the copy from step 1) will be at position 6 now
        self.MoveProperty(prop, 2)

    def MovePropertyUp(self, prop):
        """move the property one step up"""
        # here the step is -1. For example, prop is at position 5, to move it
        # to position 4, we can say move it in front of position 4. Delete the
        # original prop will not affect the position of the new copy.
        self.MoveProperty(prop, -1)

    def SendPropEvent(self, event, prop=None):
        """send the property event to the parent"""
        prop = self.GetProperty(prop)
        # prepare the event
        if isinstance(event, PropertyEvent):
            evt = event
        elif isinstance(event, int):
            evt = PropertyEvent(event)
            evt.SetProperty(prop)
        else:
            raise ValueError()

        evt.SetId(self.GetId())
        eventObject = self.GetParent()
        evt.SetEventObject(eventObject)
        evtHandler = eventObject.GetEventHandler()

        evtHandler.ProcessEvent(evt)
        return not evt.GetVeto()

    def NavigateProp(self, down):
        """change the selected property"""
        sel = self.GetSelection()
        # find the next visible property and activate it
        while True:
            if down:
                sel += 1
            else:
                sel -= 1

            if sel < 0 or sel >= self.GetPropCount():
                break

            prop = self._props[sel]
            if prop.IsVisible():
                self.SetSelection(sel)
                self.EnsureVisible(sel)
                break

    def PropHitTest(self, pt):
        """find the property under the mouse"""
        for i, prop in enumerate(self._props):
            prop = self._props[i]
            if not prop.IsVisible():
                continue
            if prop.GetRect().Contains(pt):
                return i
        return -1

    def LayoutAll(self, update=True):
        """layout the properties"""
        rc = self.GetClientRect()
        (w, h) = (rc.width, 1)

        self.CheckProp()
        # calculate the width and height
        for p in self._props:
            if p.IsVisible():
                sz = p.GetMinSize()
                w = max(w, sz.x)
                h = h + sz.y
        if update:
            # update the virtual size
            self.SetVirtualSize(wx.Size(w, h))

        # set the property rect
        rc = self.GetClientRect()
        w, y = max(w, rc.width), 1
        for p in self._props:
            if p.IsVisible():
                h = p.GetMinSize().y
                p.SetRect(wx.Rect(0, y, w, h))
                # let art provider update drawing regions (e.g., value rect)
                self._art.PrepareDrawRect(p)
                y += h

    def GetDrawRect(self):
        """return the drawing rect"""
        sz = self.GetClientSize()
        rc = wx.Rect(0, 0, sz.x, sz.y)

        # shift the client rectangle to take into account scrolling, converting
        # device to logical coordinates
        (rc.x, rc.y) = self.CalcUnscrolledPosition(rc.x, rc.y)

        return rc

    def CheckProp(self):
        """update the property status"""
        parent = None
        for i, prop in enumerate(self._props):
            parent = self.GetProperty(i - 1)
            # find the direct parent property
            while parent:
                if parent.GetIndent() < prop.GetIndent():
                    break
                parent = parent.GetParent()
            prop.SetParent(parent)
            if parent:
                # the parent has children now
                parent.SetHasChildren(True, True)
            # the current one does not have children yet; will be set by its
            # children
            prop.SetHasChildren(False, True)

        # show/hide the properties
        for prop in self._props:
            parent = prop.GetParent()
            if not parent:
                # always show prop without parent
                show = True
            else:
                # prop with parent depends on parent's status
                show = parent.IsExpanded() and parent.IsVisible()
            prop.SetVisible(show)

    def OnPropRefresh(self, evt):
        """refresh the property, for example, due to value changed"""
        self.SendPropEvent(evt.GetEventType(), evt.GetProperty())
        prop = evt.GetProperty()
        if prop is None:
            return
        rc = prop.GetRect()
        rc.x, rc.y = self.CalcScrolledPosition(rc.x, rc.y)
        self.RefreshRect(rc, True)

    def GetContextMenu(self, prop):
        # show the context menu
        menu = wx.Menu()
        menu.Append(self.ID_PROP_GRID_ADD_SEP, "&Add separator")
        menu.AppendCheckItem(self.ID_PROP_GRID_READ_ONLY, "&Read only")
        menu.Check(self.ID_PROP_GRID_READ_ONLY, prop.IsReadonly())

        menu.AppendSeparator()
        menu.Append(self.ID_PROP_GRID_INDENT_INS,
                    "Increase Indent\tCtrl-Right")
        menu.Append(self.ID_PROP_GRID_INDENT_DES, "Decrease Indent\tCtrl-Left")
        menu.AppendSeparator()
        menu.Append(self.ID_PROP_GRID_MOVE_UP, "Move up\tCtrl-Up")
        menu.Append(self.ID_PROP_GRID_MOVE_DOWN, "Move down\tCtrl-Down")
        menu.AppendSeparator()
        menu.Append(self.ID_PROP_GRID_DELETE, "&Delete")
        menu.AppendSeparator()
        menu.Append(self.ID_PROP_GRID_PROP, "&Properties")
        return menu

    def OnPropEventsHandler(self, evt):
        """process the property notification"""
        if not self.SendPropEvent(evt.GetEventType(), evt.GetProperty()):
            # vetoed by parent, ignore the event
            return False

        eid = evt.GetEventType()
        if eid in [
                wxEVT_PROP_COLLAPSED, wxEVT_PROP_EXPANDED, wxEVT_PROP_INDENT,
                wxEVT_PROP_RESIZE
        ]:
            self.UpdateGrid()
        elif eid == wxEVT_PROP_RIGHT_CLICK:
            menu = self.GetContextMenu(evt.GetProperty())
            cmd = PopupMenu(self, menu)
            self.OnProcessCommand(cmd, evt.GetProperty())
        return True

    def OnProcessCommand(self, eid, prop):
        """process the context menu command"""
        if not prop:
            return

        if eid == self.ID_PROP_GRID_DELETE:
            self.DeleteProperty(prop)
        elif eid == self.ID_PROP_GRID_READ_ONLY:
            prop.SetReadonly(not prop.IsReadonly())
        elif eid == self.ID_PROP_GRID_PROP:
            dlg = PropSettings(self, prop)
            if dlg.ShowModal() == wx.ID_OK:
                # Update the prop only may not be enough, for example, if its
                # indent is changed, it may affect the previous prop.
                self.UpdateGrid()
        elif eid == self.ID_PROP_GRID_INDENT_INS:
            prop.SetIndent(prop.GetIndent() + 1)
        elif eid == self.ID_PROP_GRID_INDENT_DES:
            prop.SetIndent(prop.GetIndent() - 1)
        elif eid == self.ID_PROP_GRID_MOVE_UP:
            self.MoveProperty(prop, -1)
        elif eid == self.ID_PROP_GRID_MOVE_DOWN:
            self.MoveProperty(prop, 2)
        elif eid == self.ID_PROP_GRID_ADD_SEP:
            self.InsertSeparator(index=self.GetSelection())

    def OnKeyDown(self, evt):
        """key down event"""
        prop = self.prop_selected
        skip = True
        if prop:
            skip = False
            index = self.GetSelection()
            keycode = evt.GetKeyCode()
            indent = prop.GetIndent()
            if keycode == wx.WXK_LEFT:
                if evt.CmdDown():
                    # Ctrl + Left decrease the indent
                    prop.SetIndent(indent - 1)
                else:
                    # Left hide children
                    prop.SetExpand(False)
            elif keycode == wx.WXK_UP:
                if evt.CmdDown():
                    # Ctrl + Up move up
                    self.MovePropertyUp(index)
                else:
                    # Up select the above property
                    self.NavigateProp(False)
            elif keycode == wx.WXK_RIGHT:
                if evt.CmdDown():
                    # Ctrl + Right increase the indent
                    prop.SetIndent(indent + 1)
                else:
                    # Right show children
                    prop.SetExpand(True)
            elif keycode == wx.WXK_DOWN:
                if evt.CmdDown():
                    # Ctrl + Down move the property down
                    self.MovePropertyDown(index)
                else:
                    # Down select the property below
                    self.NavigateProp(True)
            elif keycode == wx.WXK_DELETE:
                # delete the property
                self.RemoveProperty(self.GetSelectedProperty())
            else:
                skip = True
        if skip:
            evt.Skip()

    def OnPaint(self, event):
        """draw the property"""
        dc = wx.AutoBufferedPaintDC(self)
        self.DoPrepareDC(dc)

        rc = self.GetDrawRect()
        #draw background
        bg = self.GetBackgroundColour()
        if not bg.IsOk():
            bg = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
        pen = wx.Pen(wx.BLACK, 1, wx.PENSTYLE_TRANSPARENT)
        dc.SetPen(pen)
        brush = wx.Brush(bg)
        dc.SetBrush(brush)
        dc.DrawRectangle(rc.x, rc.y, rc.width, rc.height)

        # draw the top edge
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rc.left, rc.top, rc.right, rc.top)

        # draw the properties
        for p in self._props:
            if not p.IsVisible():
                continue
            rc_prop = p.GetRect()
            if rc.Intersects(rc_prop):
                self._art.DrawItem(dc, p)

    def OnSize(self, evt):
        """resize the properties"""
        self.UpdateGrid()
        evt.Skip()

    def OnEraseBackground(self, evt):
        """redraw the background"""
        #intentionally leave empty to remove the screen flash
        pass

    def OnMouseDown(self, evt):
        """left mouse down"""
        # find the property under mouse
        pt = self.CalcUnscrolledPosition(evt.GetPosition())
        index = self.PropHitTest(pt)
        self.pos_mouse_down = pt
        if index != -1:
            prop = self.GetProperty(index)

            # pass the event to the property
            ht = prop.OnMouseDown(pt)
            self.prop_under_mouse = prop
            self.CaptureMouse()
            self.resize_mode = self.RESIZE_NONE
            if ht == 'splitter':
                # drag the splitter
                self.resize_mode = self.RESIZE_SEP
            elif ht == 'bottom_edge':
                # drag the bottom edge
                self.resize_mode = self.RESIZE_BOT
            elif ht == 'top_edge':
                # drag the bottom edge of the property above
                if index > 0:
                    index = index - 1
                    self.prop_under_mouse = self.GetProperty(index)
                    self.resize_mode = self.RESIZE_BOT
            elif ht == 'label':
                # start drag & drop
                PropGrid.drag_start = self.ClientToScreen(pt)
                PropGrid.drag_prop = prop
                PropGrid.drag_state = 1
        # activate the property under mouse
        self.SetSelection(index)
        evt.Skip()

    def OnMouseUp(self, evt):
        """left mouse up"""
        if self.prop_under_mouse:
            pt = self.CalcUnscrolledPosition(evt.GetPosition())
            # pass the event to the property
            self.prop_under_mouse.OnMouseUp(pt)
            self.prop_under_mouse = None

        if self.GetCapture() == self:
            self.ReleaseMouse()

        # finish resizing
        self.pos_mouse_down = wx.Point(0, 0)
        self.resize_mode = self.RESIZE_NONE

        # finish drag & drop
        PropGrid.drag_prop = None
        PropGrid.drag_state = 0
        PropGrid.drag_start = wx.Point(0, 0)

        evt.Skip()

    def OnMouseDoubleClick(self, evt):
        """double click"""
        pt = self.CalcUnscrolledPosition(evt.GetPosition())
        index = self.PropHitTest(pt)
        if index != -1:
            # pass the event to the property
            prop = self.GetProperty(index)
            prop.OnMouseDoubleClick(pt)

        evt.Skip()

    def OnMouseCaptureLost(self, evt):
        pass

    def OnMouseMove(self, evt):
        """mouse move"""
        pt = self.CalcUnscrolledPosition(evt.GetPosition())
        index = self.PropHitTest(pt)
        prop = None
        if index != -1:
            # pass the event to the property
            prop = self.GetProperty(index)
            prop.OnMouseMove(pt)
        # drag & drop
        if evt.LeftIsDown() and PropGrid.drag_prop and\
           PropGrid.drag_state == 1:
            pt = self.ClientToScreen(pt)
            start = PropGrid.drag_start
            if (start.x - pt.x)**2 + (start.y - pt.y)**2 > 10:
                if self.SendPropEvent(wxEVT_PROP_BEGIN_DRAG, self.drag_prop):
                    # the mouse is moved, so start drag & drop
                    PropGrid.drag_state = 2
                    PropGrid.drag_pg = self
                    # start drag operation
                    propData = wx.TextDataObject(PropGrid.drag_prop.GetName())
                    source = wx.DropSource(PropGrid.drag_pg)
                    source.SetData(propData)

                    rtn = source.DoDragDrop(True)
                    if rtn == wx.DragError:
                        wx.LogError("An error occurred during drag \
                                     and drop operation")
                    elif rtn == wx.DragNone:
                        pass
                    elif rtn == wx.DragCopy:
                        pass
                    elif rtn == wx.DragMove:
                        pass
                    elif rtn == wx.DragCancel:
                        pass
                    PropGrid.drag_state = 0
                    PropGrid.drag_pg = None

        if evt.LeftIsDown() and self.prop_under_mouse:
            # resize the property
            if self.resize_mode == self.RESIZE_SEP:
                # resize the title width for all properties
                self._art.SetTitleWidth(max(evt.GetX(), 50))
                self.Refresh(False)
            elif self.resize_mode == self.RESIZE_BOT:
                # change the height for the property
                sz = self.prop_under_mouse.GetMinSize()
                sz2 = wx.Size(sz.x, sz.y)
                sz.y += (pt.y - self.pos_mouse_down.y)
                sz.y = max(sz.y, 25)
                if sz.y != sz2.y:
                    self.pos_mouse_down.x, self.pos_mouse_down.y = pt.x, pt.y
                    self.prop_under_mouse.SetMinSize(sz)
            else:
                self.prop_under_mouse.OnMouseMove(pt)
        else:
            if not evt.IsButton():
                # no button is pressed, show the tooltip
                tooltip = ""
                mode = self.CURSOR_STD

                if prop:
                    #pass the event to the property
                    ht = prop.OnMouseMove(pt)

                    # change the cursor icon
                    if ht == 'splitter':
                        mode = self.CURSOR_RESIZE_HORZ
                    elif ht == 'bottom_edge':
                        mode = self.CURSOR_RESIZE_VERT
                    elif ht == 'top_edge':
                        if index > 0:
                            mode = self.CURSOR_RESIZE_VERT
                        else:
                            mode = self.CURSOR_STD
                    if prop.GetShowLabelTips():
                        tooltip = prop.GetLabelTip()
                    elif prop.GetShowValueTips() and ht == 'value':
                        tooltip = prop.GetValueTip()
                    elif ht == 'expander':
                        tooltip = prop.GetLabelTip()
                # set the tooltip
                if self.GetToolTipText() != tooltip:
                    self.SetToolTip(tooltip)
                # set the cursor
                if mode != self.cursor_mode:
                    self.cursor_mode = mode
                    if mode == self.CURSOR_RESIZE_HORZ:
                        self.SetCursor(self.resize_cursor_horz)
                    elif mode == self.CURSOR_RESIZE_VERT:
                        self.SetCursor(self.resize_cursor_vert)
                    else:
                        self.SetCursor(wx.NullCursor)
        evt.Skip()

    def OnMouseLeave(self, evt):
        """mouse leaves the window"""
        self.SetCursor(wx.NullCursor)
        evt.Skip()

    def OnMouseRightClick(self, evt):
        """right click"""
        pt = self.CalcUnscrolledPosition(evt.GetPosition())
        index = self.PropHitTest(pt)
        # set the active property
        self.SetSelection(index)
        if index != -1:
            # pass the event to the property
            prop = self.GetProperty(index)
            prop.OnMouseRightClick(pt)

    def OnDrop(self, x, y, name):
        """drop the property"""
        pt = wx.Point(x, y)
        pt = self.CalcUnscrolledPosition(pt)
        index2 = self.PropHitTest(pt)
        prop = self.GetProperty(index2)
        # insert a property? Let the parent to determine what to do
        if PropGrid.drag_prop is None:
            dp.send('prop.drop', index=index2, prop=name, grid=self)
            return

        if name != PropGrid.drag_prop.GetName():
            # something is wrong
            return

        index = PropGrid.drag_pg.FindPropertyIndex(PropGrid.drag_prop)
        if index == -1:
            # not find the dragged property, do nothing
            return

        if PropGrid.drag_pg != self:
            # drop the property from the other window, copy it
            indent = PropGrid.drag_prop.GetIndent()
            self.CopyProperty(PropGrid.drag_prop, index2)
            for i in six.moves.range(index + 1,
                                     PropGrid.drag_pg.GetPropCount()):
                # copy all its children
                child = PropGrid.drag_pg.GetProperty(i)
                if child.GetIndent() <= indent:
                    break
                if index2 != -1:
                    index2 = index2 + 1
                self.CopyProperty(child, index2)
        else:
            # move the property if necessary
            if prop == PropGrid.drag_prop:
                return
            self.doMoveProperty(index, index2)
        self.UpdateGrid()


class PropSettings(wx.Dialog):
    def __init__(self, parent, prop):
        wx.Dialog.__init__(self,
                           parent,
                           title="Settings...",
                           size=wx.Size(600, 460),
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.propgrid = PropGrid(self)
        self.propgrid.GetArtProvider().SetTitleWidth(200)
        self.prop = prop
        if prop.IsSeparator():
            self.items = (('name', 'Name', '', 'editbox'), ('label', 'Label',
                                                            '', 'editbox'),
                          ('indent', 'Indent level', '', 'spin'),
                          ('font_label', 'Label font', '_font_label', 'font'),
                          ('font_value', 'Value font', '_font_value', 'font'))
        else:
            self.items = (('name', 'Name', '', 'editbox'), ('label', 'Label',
                                                            '', 'editbox'),
                          ('indent', 'Indent level', '',
                           'spin'), ('enable', 'Enable', '', 'checkbox'),
                          ('font_label', 'Label font', '_font_label',
                           'font'), ('font_value', 'Value font', '_font_value',
                                     'font'), ('readonly', 'Read only', '',
                                               'checkbox'),
                          ('text_clr', 'Normal text color', 'text_clr',
                           'color'), ('text_clr_sel', 'Selected text color',
                                      'text_clr_sel', 'color'),
                          ('text_clr_disabled', 'Disable text color',
                           'text_clr_disabled', 'color'),
                          ('bg_clr', 'Normal background color', 'bg_clr',
                           'color'), ('bg_clr_sel',
                                      'Selected background color',
                                      'bg_clr_sel', 'color'),
                          ('bg_clr_disabled', 'Disabled background color',
                           'bg_clr_disabled', 'color'))

        for (name, label, gname, ctrl) in self.items:
            pp = self.propgrid.InsertProperty(name, label, '')
            if prop:
                v = getattr(prop, name)
            else:
                v = ""
            if v is None and gname:
                v = getattr(prop.grid.GetArtProvider(), gname)

            pp.SetLabelTip(label)
            pp.SetControlStyle(ctrl)
            if ctrl == 'checkbox':
                pp.SetValue(v)
                pp.SetFormatter(BoolFormatter())
            elif ctrl == 'color':
                pp.SetValue(v)
                pp.SetBgColor(v, v, v)
                t = wx.Colour(v)
                t.SetRGB(t.GetRGB() ^ 0xffffff)
                t = t.GetAsString(wx.C2S_HTML_SYNTAX)
                pp.SetTextColor(t, t, t)
                pp.SetFormatter(ColorFormatter())
            elif ctrl == 'font':
                pp.SetValueFont(v)
                pp.SetValue(v)
                pp.SetFormatter(FontFormatter())
            elif ctrl in ['spin', 'slider']:
                pp.SetFormatter(IntFormatter(0, 100))
                pp.SetValue(v)
            else:
                pp.SetValue(v)

        sz = wx.BoxSizer(wx.VERTICAL)
        sz.Add(self.propgrid, 1, wx.EXPAND | wx.ALL, 1)

        self.stcline = wx.StaticLine(self, style=wx.LI_HORIZONTAL)
        sz.Add(self.stcline, 0, wx.EXPAND | wx.ALL, 5)

        sz2 = wx.BoxSizer(wx.HORIZONTAL)
        sz2.Add(wx.Button(self, wx.ID_OK, u"&Ok"), 0, wx.ALL, 5)
        sz2.Add(wx.Button(self, wx.ID_CANCEL, u"&Cancel"), 0, wx.ALL, 5)
        sz.Add(sz2, 0, wx.ALIGN_RIGHT | wx.RIGHT, 5)

        self.SetSizer(sz)
        self.Layout()

        # Connect Events
        self.Bind(wx.EVT_BUTTON, self.OnBtnOk, id=wx.ID_OK)
        self.Bind(EVT_PROP_RIGHT_CLICK, self.OnPropEventsHandler)
        self.Bind(EVT_PROP_CHANGED, self.OnPropChanged)

    def OnPropEventsHandler(self, evt):
        # disable context menu
        evt.Veto()

    def OnPropChanged(self, evt):
        p = evt.GetProperty()
        if '_clr' in p.GetName():
            t = wx.Colour(p.GetValue())
            c = t.GetAsString(wx.C2S_HTML_SYNTAX)
            p.SetBgColor(c, c, c)
            t.SetRGB(t.GetRGB() ^ 0xFFFFFF)
            t = t.GetAsString(wx.C2S_HTML_SYNTAX)
            p.SetTextColor(t, t, t)
        elif 'font_' in p.GetName():
            p.SetValueFont(p.GetValue())

    def OnBtnOk(self, event):
        if self.propgrid.prop_selected:
            # update the value if necessary
            self.propgrid.prop_selected.OnTextEnter()

        for (name, _, _, ctrl) in self.items:
            v = self.propgrid.GetProperty(name)
            if ctrl == 'checkbox':
                setattr(self.prop, name, bool(int(v.GetValue())))
            if getattr(self.prop, name) is None:
                setattr(self.prop, name, v.GetValue())
            else:
                setattr(self.prop, name,
                        type(getattr(self.prop, name))(v.GetValue()))
        event.Skip()



import types
import sys

if sys.version_info.major == 2:
    def rebound_instance_method(f, newobj):
        return types.MethodType(f.im_func, newobj, newobj.__class__)
else:
    def rebound_instance_method(f, newobj):
        return lambda *args, **kwargs: f.__func__(newobj, *args, **kwargs)

if sys.version_info.major == 2:
    def generic_class_method(f):
        return f.im_func
else:
    def generic_class_method(f):
        return f

if sys.version_info.major >= 3:
    unicode = str


def setup_Structure_debug_helper():
    import ctypes

    class DebugStruct(ctypes.Structure):
        def __init__(self, *args):
            assert hasattr(self, "_fields_")
            assert len(args) <= len(self._fields_)
            super(DebugStruct, self).__init__(*args)

    ctypes.Structure = DebugStruct



import multiprocessing
import wx
import click
from .mainframe import MainFrame
from .version import __version__

class RunApp(wx.App):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        wx.App.__init__(self, redirect=False)

    def OnInit(self):
        wx.Log.SetActiveTarget(wx.LogStderr())

        self.SetAssertMode(wx.APP_ASSERT_DIALOG)

        frame = MainFrame(None, **self.kwargs)
        frame.Show(True)
        self.SetTopWindow(frame)
        self.frame = frame
        return True


@click.command()
@click.version_option(__version__)
@click.option('--config',
              '-c',
              default='bsmedit',
              help="Set configuration file name, default 'bsmedit'.")
@click.option('--path',
              '-p',
              multiple=True,
              type=click.Path(exists=True),
              help="Add external module path.")
@click.option('--ignore-perspective',
              '-i',
              is_flag=True,
              help="Do not load perspective.")
@click.option('--spawn',
              is_flag=True,
              help="Start a process with method 'spawn'.")
@click.argument('module', nargs=-1)
def main(config, path, ignore_perspective, spawn, module):
    if spawn and hasattr(multiprocessing, 'set_start_method'):
        multiprocessing.set_start_method('spawn')
    app = RunApp(config=config,
                 ignore_perspective=ignore_perspective,
                 path=path,
                 module=module)
    app.MainLoop()


if __name__ == '__main__':
    main()




import ast


# We could use this funny little hack:
# background: http://stackoverflow.com/questions/6959360/goto-in-python
# code from here: http://code.activestate.com/recipes/576944-the-goto-decorator/
# by Carl Cerecke
# Licensed under the MIT License

class MissingLabelError(Exception):
    """'goto' without matching 'label'."""
    pass

def goto(fn):
    import dis
    import new

    """
    A function decorator to add the goto command for a function.

    Specify labels like so:

    label .foo

    Goto labels like so:

    goto .foo
    """
    labels = {}
    gotos = {}
    globalName = None
    index = 0
    end = len(fn.func_code.co_code)
    i = 0

    # scan through the byte codes to find the labels and gotos
    while i < end:
        op = ord(fn.func_code.co_code[i])
        i += 1
        name = dis.opname[op]

        if op > dis.HAVE_ARGUMENT:
            b1 = ord(fn.func_code.co_code[i])
            b2 = ord(fn.func_code.co_code[i+1])
            num = b2 * 256 + b1

            if name == 'LOAD_GLOBAL':
                globalName = fn.func_code.co_names[num]
                index = i - 1
                i += 2
                continue

            if name == 'LOAD_ATTR':
                if globalName == 'label':
                    labels[fn.func_code.co_names[num]] = index
                elif globalName == 'goto':
                    gotos[fn.func_code.co_names[num]] = index

            name = None
            i += 2

    # no-op the labels
    ilist = list(fn.func_code.co_code)
    for label,index in labels.items():
        ilist[index:index+7] = [chr(dis.opmap['NOP'])]*7

    # change gotos to jumps
    for label,index in gotos.items():
        if label not in labels:
            raise MissingLabelError("Missing label: %s"%label)

        target = labels[label] + 7   # skip NOPs
        ilist[index] = chr(dis.opmap['JUMP_ABSOLUTE'])
        ilist[index + 1] = chr(target & 255)
        ilist[index + 2] = chr(target >> 8)

    # create new function from existing function
    c = fn.func_code
    newcode = new.code(c.co_argcount,
                       c.co_nlocals,
                       c.co_stacksize,
                       c.co_flags,
                       ''.join(ilist),
                       c.co_consts,
                       c.co_names,
                       c.co_varnames,
                       c.co_filename,
                       c.co_name,
                       c.co_firstlineno,
                       c.co_lnotab)
    newfn = new.function(newcode,fn.func_globals)
    return newfn


# Or, to avoid messing around with the bytecode, we can go the
# much more complicated route:
# - First, we flatten any Python AST into a series of statements,
#   with even more goto's added for while/for loops, if's, etc.
#   (We can avoid doing this for all sub-ASTs where there are no
#    no goto-labels. The goal is to have all goto-labels at
#    the top-level, not inside a sub-AST.)
# - Only conditional goto's can stay.
# - All Gotos and Goto-labels are marked somehow as special elements.
# So, we end up with sth like:
#   x()
#   y()
#   <goto-label "a">
#   z()
#   <goto-stmnt "a">
#   w()
#   if v(): <goto-stmnt "a">
#   q()
# Now, we can implement the goto-handling based on this flattened code:
# - Add a big endless loop around it. After the final statement,
#   a break would leave the loop.
# - Before the loop, we add the statement `goto = None`.
# - The goto-labels will split the code into multiple part, where
#   we add some `if goto is None:` before each part
#   (excluding the goto-labels).
# - For the goto-labels itself, we add this code:
#   `if goto == <goto-label>: goto = None`
# - For every goto-statement, we add this code:
#   `goto = <goto-stmnt>; continue`


class GotoLabel:
    def __init__(self, label):
        self.label = label


class GotoStatement:
    def __init__(self, label):
        self.label = label


class _Flatten:
    def __init__(self):
        self.c = 1

    def make_jump(self):
        label = self.c
        self.c += 1
        return GotoStatement(label)

    def flatten(self, body, breakJump=None):
        """
        :type body: list[ast.AST]
        :param breakJump: if we find some ast.Break in a while-loop, add this jump
        :rtype: list[ast.AST]
        """
        r = []
        for s in body:
            if isinstance(s, ast.If):
                a = ast.UnaryOp()
                a.op = ast.Not()
                a.operand = s.test
                goto_final_stmnt = self.make_jump()
                if s.orelse:
                    goto_orelse_stmnt = self.make_jump()
                    r += [ast.If(test=a, body=[goto_orelse_stmnt], orelse=[])]
                else:
                    goto_orelse_stmnt = None
                    r += [ast.If(test=a, body=[goto_final_stmnt], orelse=[])]
                r += self.flatten(s.body, breakJump=breakJump)
                if s.orelse:
                    r += [goto_final_stmnt]
                    r += [GotoLabel(goto_orelse_stmnt.label)]
                    r += self.flatten(s.orelse, breakJump=breakJump)
                r += [GotoLabel(goto_final_stmnt.label)]
            elif isinstance(s, ast.While):
                if s.orelse: raise NotImplementedError
                goto_repeat_stmnt = self.make_jump()
                r += [GotoLabel(goto_repeat_stmnt.label)]
                a = ast.UnaryOp()
                a.op = ast.Not()
                a.operand = s.test
                goto_final_stmnt = self.make_jump()
                r += [ast.If(test=a, body=[goto_final_stmnt], orelse=[])]
                r += self.flatten(s.body, breakJump=goto_final_stmnt)
                r += [goto_repeat_stmnt]
                r += [GotoLabel(goto_final_stmnt.label)]
            elif isinstance(s, ast.For):
                raise NotImplementedError
            elif isinstance(s, (ast.TryExcept, ast.TryFinally)):
                raise NotImplementedError
            elif isinstance(s, ast.Break):
                assert breakJump, "found break in unexpected scope"
                r += [breakJump]
            else:
                r += [s]
        return r


def _ast_for_value(v):
    if isinstance(v, (str,unicode)): return ast.Str(s=v)
    elif isinstance(v, int): return ast.Num(n=v)
    else: raise NotImplementedError("type (%r) %r" % (type(v), v))


class _HandleGoto:

    def __init__(self, gotoVarName):
        self.gotoVarName = gotoVarName

    def handle_goto_stmnt(self, stmnt):
        assert isinstance(stmnt, GotoStatement)
        a = ast.Assign()
        a.targets = [ast.Name(id=self.gotoVarName, ctx=ast.Store())]
        a.value = _ast_for_value(stmnt.label)
        return [a, ast.Continue()]

    def handle_goto_label(self, stmnt):
        assert isinstance(stmnt, GotoLabel)
        reset_ast = ast.Assign()
        reset_ast.targets = [ast.Name(id=self.gotoVarName, ctx=ast.Store())]
        reset_ast.value = ast.Name(id="None", ctx=ast.Load())
        test_ast = ast.Compare()
        test_ast.ops = [ast.Eq()]
        test_ast.left = ast.Name(id=self.gotoVarName, ctx=ast.Store())
        test_ast.comparators = [_ast_for_value(stmnt.label)]
        return [ast.If(test=test_ast, body=[reset_ast], orelse=[])]

    def handle_body(self, body):
        """
        :type body: list[ast.AST]
        :rtype: list[ast.AST]
        """
        parts = [[]]
        for s in body:
            if isinstance(s, GotoLabel):
                parts += [s, []]
            else:
                parts[-1].append(s)
        r = []
        for l in parts:
            if not l: continue
            if isinstance(l, GotoLabel):
                r += self.handle_goto_label(l)
            else:
                sr = []
                for s in l:
                    if isinstance(s, ast.If):
                        assert not s.orelse
                        assert len(s.body) == 1
                        assert isinstance(s.body[0], GotoStatement)
                        sr += [ast.If(test=s.test, orelse=[],
                                      body=self.handle_goto_stmnt(s.body[0]))]
                    elif isinstance(s, (ast.While, ast.For)):
                        assert False, "not expected: %r" % s
                    elif isinstance(s, GotoStatement):
                        sr += self.handle_goto_stmnt(s)
                    else:
                        sr += [s]
                test_ast = ast.Compare()
                test_ast.ops = [ast.Is()]
                test_ast.left = ast.Name(id=self.gotoVarName, ctx=ast.Store())
                test_ast.comparators = [ast.Name(id="None", ctx=ast.Load())]
                r += [ast.If(test=test_ast, body=sr, orelse=[])]
        return r

    def wrap_func_body(self, flat_body):
        var_ast = ast.Assign()
        var_ast.targets = [ast.Name(id=self.gotoVarName, ctx=ast.Store())]
        var_ast.value = ast.Name(id="None", ctx=ast.Load())
        main_loop_ast = ast.While(orelse=[])
        main_loop_ast.test = ast.Name(id="True", ctx=ast.Load())
        main_loop_ast.body = self.handle_body(flat_body)
        main_loop_ast.body += [ast.Break()]
        return [var_ast, main_loop_ast]


def transform_goto(f, gotoVarName):
    assert isinstance(f, ast.FunctionDef)
    flat_body = _Flatten().flatten(f.body)
    new_body = _HandleGoto(gotoVarName).wrap_func_body(flat_body)
    new_func_ast = ast.FunctionDef(
        name=f.name,
        args=f.args,
        decorator_list=f.decorator_list,
        body=new_body)
    return new_func_ast



# PyCParser - global include wrappers
# by Albert Zeyer, 2011
# code under BSD 2-Clause License

from cparser import *
from interpreter import CWrapValue, _ctype_ptr_get_value, Helpers
import ctypes, _ctypes
import errno, os

libc = ctypes.CDLL(None)

def _fixCType(stateStruct, t):
    if t is ctypes.c_void_p: t = CBuiltinType(("void", "*"))
    if t is ctypes.c_char_p: t = CPointerType(CBuiltinType(("char",)))
    if t is ctypes.c_char: t = CBuiltinType(("char",))
    return t

def wrapCFunc(state, funcname, restype, argtypes, varargs=False):
    f = getattr(libc, funcname)
    restype = _fixCType(state, restype)
    if restype is CVoidType:
        f.restype = None
    else:
        assert restype is not None
        f.restype = getCTypeWrapped(restype, state)
    assert argtypes is not None
    argtypes = [_fixCType(state, arg) for arg in argtypes]
    f.argtypes = [getCTypeWrapped(arg, state) for arg in argtypes]
    state.funcs[funcname] = CWrapValue(
        f, name=funcname, funcname=funcname,
        returnType=restype, argTypes=argtypes)

def wrapCFunc_varargs(state, funcname, wrap_funcname):
    """
    :param str funcname: e.g. "vprintf"
    :param wrap_funcname: e.g. "printf"
    Will register a new function, where the last arg is expected to be va_list.
    va_list is just a tuple of args.
    Will call the wrap-func with all args and unwraps the va_list args.
    """
    wrap_func = state.funcs[wrap_funcname]
    assert isinstance(wrap_func, CWrapValue)
    wrap_arg_len = len(wrap_func.value.argtypes)
    def f(*args):
        assert len(args) == wrap_arg_len + 1
        assert isinstance(args[-1], Helpers.VarArgs)
        return wrap_func.value(*(args[:-1] + args[-1].args))
    f.__name__ = funcname
    state.funcs[funcname] = CWrapValue(
        f, name=funcname, funcname=funcname,
        returnType=wrap_func.returnType, argTypes=wrap_func.argTypes)

def _fixCArg(a):
    if isinstance(a, unicode):
        a = a.encode("utf-8")
    if isinstance(a, str):
        a = ctypes.c_char_p(a)
    if isinstance(a, ctypes.c_char_p) or (isinstance(a, _ctypes._Pointer) and a._type_ is ctypes.c_char):
        return ctypes.cast(a, ctypes.POINTER(ctypes.c_byte))
    if isinstance(a, ctypes.c_char):
        return ctypes.c_byte(ord(a.value))
    return a

def callCFunc(funcname, *args):
    f = getattr(libc, funcname)
    args = map(_fixCArg, args)
    return f(*args)


class Wrapper:
    def __init__(self, state):
        """
        :type state: cparser.State
        """
        self.state = state
        # The Wrapper is supposed to work for parsing also without an interpreter.
        # However, when you are going to call some of the functions from here,
        # this is needed.
        self.interpreter = None

    def handle_limits_h(self, state):
        state.macros["UCHAR_MAX"] = Macro(rightside="255")
        state.macros["INT_MAX"] = Macro(rightside=str(2 ** (ctypes.sizeof(ctypes.c_int) * 8 - 1)))
        state.macros["ULONG_MAX"] = Macro(rightside=str(2 ** (ctypes.sizeof(ctypes.c_ulong) * 8) - 1))
    def handle_stdio_h(self, state):
        state.macros["NULL"] = Macro(rightside="0")
        FileP = CPointerType(CStdIntType("FILE")).getCType(state)
        wrapCFunc(state, "fopen", restype=FileP, argtypes=(ctypes.c_char_p, ctypes.c_char_p))
        wrapCFunc(state, "fclose", restype=ctypes.c_int, argtypes=(FileP,))
        wrapCFunc(state, "fdopen", restype=FileP, argtypes=(ctypes.c_int, ctypes.c_char_p))
        state.vars["stdin"] = CWrapValue(callCFunc("fdopen", 0, "r"), name="stdin")
        state.vars["stdout"] = CWrapValue(callCFunc("fdopen", 1, "a"), name="stdout")
        state.vars["stderr"] = CWrapValue(callCFunc("fdopen", 2, "a"), name="stderr")
        wrapCFunc(state, "printf", restype=ctypes.c_int, argtypes=(ctypes.c_char_p,), varargs=True)
        wrapCFunc(state, "fprintf", restype=ctypes.c_int, argtypes=(FileP, ctypes.c_char_p), varargs=True)
        wrapCFunc(state, "sprintf", restype=ctypes.c_int, argtypes=(ctypes.c_char_p, ctypes.c_char_p), varargs=True)
        wrapCFunc_varargs(state, "vprintf", wrap_funcname="printf")
        wrapCFunc_varargs(state, "vfprintf", wrap_funcname="fprintf")
        wrapCFunc_varargs(state, "vsprintf", wrap_funcname="sprintf")
        wrapCFunc(state, "fputs", restype=ctypes.c_int, argtypes=(ctypes.c_char_p, FileP))
        wrapCFunc(state, "fputc", restype=ctypes.c_int, argtypes=(ctypes.c_int, FileP))
        wrapCFunc(state, "fgets", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p, ctypes.c_int, FileP))
        wrapCFunc(state, "fread", restype=ctypes.c_size_t, argtypes=(ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, FileP))
        wrapCFunc(state, "fwrite", restype=ctypes.c_size_t, argtypes=(ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, FileP))
        wrapCFunc(state, "fflush", restype=ctypes.c_int, argtypes=(FileP,))
        wrapCFunc(state, "ftell", restype=ctypes.c_long, argtypes=(FileP,))
        wrapCFunc(state, "rewind", restype=CVoidType, argtypes=(FileP,))
        wrapCFunc(state, "ferror", restype=ctypes.c_int, argtypes=(FileP,))
        wrapCFunc(state, "clearerr", restype=CVoidType, argtypes=(FileP,))
        state.vars["errno"] = CWrapValue(0, name="errno") # TODO
        state.macros["EOF"] = Macro(rightside="-1") # TODO?
        wrapCFunc(state, "setbuf", restype=CVoidType, argtypes=(FileP, ctypes.c_char_p))
        wrapCFunc(state, "isatty", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "fileno", restype=ctypes.c_int, argtypes=(FileP,))
        wrapCFunc(state, "getc", restype=ctypes.c_int, argtypes=(FileP,))
        wrapCFunc(state, "ungetc", restype=ctypes.c_int, argtypes=(ctypes.c_int,FileP))
        struct_stat = state.structs["stat"] = CStruct(name="stat") # TODO
        struct_stat.body = CBody(parent=struct_stat)
        CVarDecl(parent=struct_stat, name="st_mode", type=ctypes.c_int).finalize(state)
        state.funcs["fstat"] = CWrapValue(lambda *args: None, returnType=ctypes.c_int, name="fstat") # TODO
        state.macros["S_IFMT"] = Macro(rightside="0") # TODO
        state.macros["S_IFDIR"] = Macro(rightside="0") # TODO
    def handle_stdlib_h(self, state):
        state.macros["EXIT_SUCCESS"] = Macro(rightside="0")
        state.macros["EXIT_FAILURE"] = Macro(rightside="1")
        state.funcs["abort"] = CWrapValue(
            lambda: self.interpreter._abort(),
            returnType=CVoidType,
            name="abort"
        )
        state.funcs["exit"] = CWrapValue(
            lambda s: self.interpreter._exit(s.value),  # int
            returnType=CVoidType,
            name="exit"
        )
        state.funcs["malloc"] = CWrapValue(
            lambda s: self.interpreter._malloc(s.value),  # size_t
            returnType=ctypes.c_void_p,
            name="malloc"
        )
        state.funcs["realloc"] = CWrapValue(
            lambda (p, s): self.interpreter._realloc(_ctype_ptr_get_value(p), s.value),  # void*, size_t
            returnType=ctypes.c_void_p,
            name="realloc"
        )
        state.funcs["free"] = CWrapValue(
            lambda p: self.interpreter._free(_ctype_ptr_get_value(p)),  # void*
            returnType=CVoidType,
            name="free"
        )
        wrapCFunc(state, "strtoul", restype=ctypes.c_ulong, argtypes=(ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int))
        state.funcs["atoi"] = CWrapValue(
            lambda x: ctypes.c_int(int(ctypes.cast(x, ctypes.c_char_p).value)),
            returnType=ctypes.c_int,
            name="atoi"
        )
        state.funcs["getenv"] = CWrapValue(
            lambda x: _fixCArg(ctypes.c_char_p(os.getenv(ctypes.cast(x, ctypes.c_char_p).value))),
            returnType=CPointerType(ctypes.c_byte),
            name="getenv"
        )
    def handle_stdarg_h(self, state):
        state.typedefs["va_list"] = CTypedef(name="va_list", type=CVariadicArgsType())
        def va_start(v, dummy_last):
            assert isinstance(v, Helpers.VarArgs)
            v.idx = 0
        def va_end(v):
            assert isinstance(v, Helpers.VarArgs)
            #assert v.idx == len(v.args), "VarArgs: va_end: not handled all args"  # is this an error?
        def __va_arg(v, inplace_typed):
            assert isinstance(v, Helpers.VarArgs)
            x = v.get_next()
            helpers = v.intp.helpers
            helpers.assignGeneric(inplace_typed, x)
            return inplace_typed
        def __va_arg_getReturnType(stateStruct, stmnt_args):
            assert len(stmnt_args) == 2  # see __va_arg
            return getValueType(stateStruct, stmnt_args[1])
        state.funcs["va_start"] = CWrapValue(va_start, name="va_start", returnType=CVoidType)
        state.funcs["va_end"] = CWrapValue(va_end, name="va_end", returnType=CVoidType)
        state.macros["va_arg"] = Macro(args=("list", "type"), rightside="((__va_arg(list, type())))")
        state.funcs["__va_arg"] = CWrapValue(__va_arg, name="__va_arg",
                                             returnType=None, getReturnType=__va_arg_getReturnType)
    def handle_stddef_h(self, state): pass
    def handle_math_h(self, state): pass
    def handle_string_h(self, state):
        wrapCFunc(state, "strlen", restype=ctypes.c_size_t, argtypes=(ctypes.c_char_p,))
        wrapCFunc(state, "strcpy", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_char_p))
        wrapCFunc(state, "strncpy", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_char_p,ctypes.c_size_t))
        wrapCFunc(state, "strcat", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_char_p))
        wrapCFunc(state, "strcmp", restype=ctypes.c_int, argtypes=(ctypes.c_char_p,ctypes.c_char_p))
        wrapCFunc(state, "strncmp", restype=ctypes.c_int, argtypes=(ctypes.c_char_p,ctypes.c_char_p,ctypes.c_size_t))
        wrapCFunc(state, "strtok", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_char_p))
        wrapCFunc(state, "strchr", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_int))
        wrapCFunc(state, "strrchr", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_int))
        wrapCFunc(state, "strstr", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,ctypes.c_char_p))
        wrapCFunc(state, "strdup", restype=ctypes.c_char_p, argtypes=(ctypes.c_char_p,))
        wrapCFunc(state, "strerror", restype=ctypes.c_char_p, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "memset", restype=ctypes.c_void_p, argtypes=(ctypes.c_void_p, ctypes.c_int, ctypes.c_size_t))
        wrapCFunc(state, "memcpy", restype=ctypes.c_void_p, argtypes=(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t))
        wrapCFunc(state, "memmove", restype=ctypes.c_void_p, argtypes=(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t))
        wrapCFunc(state, "memchr", restype=ctypes.c_void_p, argtypes=(ctypes.c_void_p, ctypes.c_int, ctypes.c_size_t))
        wrapCFunc(state, "memcmp", restype=ctypes.c_int, argtypes=(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t))
    def handle_time_h(self, state): pass
    def handle_ctype_h(self, state):
        wrapCFunc(state, "isalpha", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "isalnum", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "isspace", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "isdigit", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "isxdigit", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "islower", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "tolower", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "isupper", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
        wrapCFunc(state, "toupper", restype=ctypes.c_int, argtypes=(ctypes.c_int,))
    def handle_wctype_h(self, state): pass
    def handle_assert_h(self, state):
        def assert_wrap(x):
            if isinstance(x, (ctypes._Pointer, ctypes.Array, ctypes._CFuncPtr)):
                x = ctypes.cast(x, ctypes.c_void_p)
            assert x.value
        state.funcs["assert"] = CWrapValue(assert_wrap, returnType=CVoidType, name="assert")
    def handle_fcntl_h(self, state):
        state.macros["O_RDONLY"] = Macro(rightside="0x0000")
        wrapCFunc(state, "open", restype=ctypes.c_int, argtypes=(ctypes.c_char_p, ctypes.c_int))
        wrapCFunc(state, "read", restype=ctypes.c_int, argtypes=(ctypes.c_int, ctypes.c_void_p, ctypes.c_size_t))  # normally <unistd.h>
        wrapCFunc(state, "close", restype=ctypes.c_int, argtypes=(ctypes.c_int,))  # normally <unistd.h>
        # TODO: these are on OSX. cross-platform? probably not...
        state.macros["EINTR"] = Macro(rightside="4")  # via <sys/errno.h>
        state.macros["ERANGE"] = Macro(rightside="34")  # via <sys/errno.h>
    def handle_signal_h(self, state):
        # typedef void (*sig_t) (int)
        state.typedefs["sig_t"] = CTypedef(
            name="sig_t", type=CFuncPointerDecl(type=CVoidType(), args=[CBuiltinType(("int",))]))
        # There is no safe way to support the native C function.
        # The signal handler can be called at any point and it could be that
        # the GIL is hold. Then the signal handler code deadlocks because it also wants the GIL.
        #wrapCFunc(state, "signal", restype=state.typedefs["sig_t"],
        #          argtypes=(ctypes.c_int, state.typedefs["sig_t"]))
        def signal(sig, f):
            sig = sig.value
            import signal
            if isinstance(f, CWrapValue):
                f = f.value
            def sig_handler(sig, stack_frame):
                return f(sig)
            if isinstance(f, ctypes._CFuncPtr):
                if _ctype_ptr_get_value(f) == 0:  # place-holder for SIG_DFL
                    sig_handler = signal.SIG_DFL
                elif _ctype_ptr_get_value(f) == 1:  # place-holder for SIG_IGN
                    sig_handler = signal.SIG_IGN
            old_action = signal.signal(sig, sig_handler)
            # TODO: need to use helpers.makeFuncPtr for old_action.
            # And maybe handle SIG_DFL/SIG_IGN cases?
            return 0  # place-holder for SIG_DFL
        state.funcs["signal"] = CWrapValue(signal, name="signal", returnType=state.typedefs["sig_t"])
        state.macros["SIGINT"] = Macro(rightside="2")
        state.macros["SIG_DFL"] = Macro(rightside="((sig_t)0)")
        state.macros["SIG_IGN"] = Macro(rightside="((sig_t)1)")
        state.macros["SIG_ERR"] = Macro(rightside="((sig_t)-1)")
    def handle_locale_h(self, state):
        struct_lconv = state.structs["lconv"] = CStruct(name="lconv") # TODO
        struct_lconv.body = CBody(parent=struct_lconv)
        CVarDecl(parent=struct_lconv, name="grouping", type=ctypes.c_char_p).finalize(state)
        CVarDecl(parent=struct_lconv, name="thousands_sep", type=ctypes.c_char_p).finalize(state)
        wrapCFunc(state, "localeconv", restype=struct_lconv, argtypes=())
    def handle_sys_types_h(self, state):
        pass  # dummy

    def find_handler_func(self, filename):
        funcname = "handle_" + filename.replace("/", "_").replace(".", "_")
        return getattr(self, funcname, None)

    def readGlobalInclude(self, state, oldFunc, filename):
        f = self.find_handler_func(filename)
        if f is not None:
            def reader():
                if filename in state._global_include_list: return  # already included
                f(state)
                state._global_include_list.append(filename)
                return
                yield None # to make it a generator
            return reader(), None
        return oldFunc(filename) # fallback

    def install(self):
        state = self.state
        oldFunc = state.readGlobalInclude
        state.readGlobalInclude = lambda fn: self.readGlobalInclude(state, oldFunc, fn)

    def add_all_to_state(self, state):
        for funcname in dir(self):
            if not funcname.startswith("handle_"): continue
            f = getattr(self, funcname)
            f(state)
